"""
The setting of Superb ASV

Authors
  * Po-Han Chi 2021
  * Leo 2021
  * Haibin Wu 2022
  * Leo 2022
"""

import logging
import pickle
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
from omegaconf import MISSING
from torch.utils.data import Subset

from s3prl.dataio.corpus.voxceleb1sv import VoxCeleb1SV
from s3prl.dataio.dataset import EncodeCategory, LoadAudio, get_info
from s3prl.dataio.encoder.category import CategoryEncoder
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler
from s3prl.nn.speaker_model import SuperbXvector

from .run import ASV

SAMPLE_RATE = 16000
EFFECTS = [
    ["channels", "1"],
    ["rate", "16000"],
    ["gain", "-3.0"],
    ["silence", "1", "0.1", "0.1%", "-1", "0.1", "0.1%"],
]

logger = logging.getLogger(__name__)

__all__ = [
    "prepare_voxceleb1_for_sv",
    "SuperbASV",
]


def prepare_voxceleb1_for_sv(
    target_dir: str,
    cache_dir: str,
    get_path_only: str,
    dataset_root: str,
    force_download: bool = False,
):
    """
    Prepare VoxCeleb1 for speaker verification
    following :obj:`SuperbASV.prepare_data` format.

    Args:
        dataset_root (str): The root path of Fluent Speech Command
        force_download (bool): always re-download the metadata for VoxCeleb1
    """

    train_path = target_dir / "train.csv"
    test_trial_path = target_dir / "test_trial.csv"

    if get_path_only:
        return train_path, [test_trial_path]

    corpus = VoxCeleb1SV(dataset_root, cache_dir, force_download)
    train_data, valid_data, test_data, test_trials = corpus.all_data
    all_data = {**train_data, **valid_data}

    ids = sorted(all_data.keys())
    wav_paths = [all_data[idx]["wav_path"] for idx in ids]
    labels = [all_data[idx]["label"] for idx in ids]
    pd.DataFrame({"id": ids, "wav_path": wav_paths, "spk": labels}).to_csv(
        train_path, index=False
    )

    labels, id1s, id2s = zip(*test_trials)
    wav_path1 = [test_data[idx]["wav_path"] for idx in id1s]
    wav_path2 = [test_data[idx]["wav_path"] for idx in id2s]
    pd.DataFrame(
        {
            "id1": id1s,
            "id2": id2s,
            "wav_path1": wav_path1,
            "wav_path2": wav_path2,
            "label": labels,
        }
    ).to_csv(test_trial_path, index=False)

    return train_path, [test_trial_path]


class SuperbASV(ASV):
    def default_config(self):
        return dict(
            target_dir=MISSING,
            cache_dir=None,
            test_ckpt_steps=None,  # eval all saved checkpoints
            prepare_data=dict(
                dataset_root=MISSING,
            ),
            build_dataset=dict(
                train=dict(
                    min_secs=2.0,
                    max_secs=8.0,
                ),
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=10,
                    shuffle=True,
                ),
                test=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                loss_type="amsoftmax",
                loss_conf=dict(
                    margin=0.4,
                    scale=30,
                ),
            ),
            build_optimizer=dict(
                name="AdamW",
                conf=dict(
                    lr=1.0e-4,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            train=dict(
                total_steps=200000,
                log_step=500,
                eval_step=1e20,
                save_step=10000,
                gradient_clipping=1.0e3,
                gradient_accumulate=5,
                valid_metric=None,
                valid_higher_better=None,
                auto_resume=True,
                resume_ckpt_dir=None,
                keep_num_ckpts=None,
            ),
        )

    def prepare_data(
        self, prepare_data: dict, target_dir: str, cache_dir: str, get_path_only: bool
    ):
        """
        Prepare the task-specific data metadata (path, labels...).
        By default call :obj:`prepare_voxceleb1_for_sv` with :code:`**prepare_data`

        Args:
            prepare_data (dict): same in :obj:`default_config`,
                support arguments in :obj:`prepare_voxceleb1_for_sv`
            target_dir (str): Parse your corpus and save the csv file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            get_path_only (bool): Directly return the filepaths no matter they exist or not.

        Returns:
            tuple

            1. train_path (str)
            2. test_trial_paths (List[str])

            The :code:`train_path` should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            id                    (str) - the unique id for this utterance
            wav_path              (str) - the absolute path of the waveform file
            spk                   (str) - a string speaker label
            ====================  ====================

            Each :code:`test_trial_path` should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            id1                   (str) - the unique id of the first utterance
            id2                   (str) - the unique id of the second utterance
            wav_path1             (str) - the absolute path of the first utterance
            wav_path2             (str) - the absolute path of the second utterance
            label                 (int) - 0 when two utterances are from different speakers, \
                                    1 when same speaker
            ====================  ====================
        """
        return prepare_voxceleb1_for_sv(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )

    def build_encoder(
        self,
        build_encoder: dict,
        target_dir: str,
        cache_dir: str,
        train_csv: str,
        test_csvs: list,
        get_path_only: bool,
    ):
        """
        Build the encoder (for the labels) given the data metadata, and return the saved encoder path.
        By default generate and save a :obj:`s3prl.dataio.encoder.CategoryEncoder` from the :code:`label` column of the train csv.

        Args:
            build_encoder (dict): same in :obj:`default_config`, no argument supported for now
            target_dir (str): Save your encoder into this directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            train_csv_path (str): the train path from :obj:`prepare_data`
            valid_csv_path (str): the valid path from :obj:`prepare_data`
            test_csv_paths (List[str]): the test paths from :obj:`prepare_data`
            get_path_only (bool): Directly return the filepaths no matter they exist or not

        Returns:
            str

            encoder_path: The encoder should be saved in the pickle format
        """
        encoder_path = Path(target_dir) / "spk2int.pkl"
        if get_path_only:
            return encoder_path

        csv = pd.read_csv(train_csv)
        all_spk = sorted(set(csv["spk"]))
        spk2int = CategoryEncoder(all_spk)

        with open(encoder_path, "wb") as f:
            pickle.dump(spk2int, f)

        return encoder_path

    def build_dataset(
        self,
        build_dataset: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        encoder_path: str,
    ):
        """
        Build the dataset for train/valid/test.

        Args:
            build_dataset (dict): same in :obj:`default_config`, have
                :code:`train` and :code:`test` keys, each is a dictionary, for :code:`train` dictionary:

                ====================  ====================
                key                   description
                ====================  ====================
                min_secs              (float) - Drop a waveform if it is not longer than :code:`min_secs`
                max_secs              (float) - If a waveform is longer than :code:`max_secs` seconds, \
                                        randomly crop the waveform into :code:`max_secs` seconds. \
                                        Default: None, no cropping
                ====================  ====================

                for :code:`test` dictionary, no argument supported yet

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): The metadata csv file for the specific :code:`mode`
            encoder_path (str): The pickled encoder path for encoding the labels

        Returns:
            torch Dataset

            For train mode, the dataset should return each item as a dictionary
            containing the following keys:

            ====================  ====================
            key                   description
            ====================  ====================
            x                     (torch.FloatTensor) - the waveform in (seq_len, 1)
            x_len                 (int) - the waveform length :code:`seq_len`
            class_id              (str) - the label class id encoded by :code:`encoder_path`
            unique_name           (str) - the unique id for this datapoint
            ====================  ====================

            For test mode:

            ====================  ====================
            key                   description
            ====================  ====================
            x                     (torch.FloatTensor) - the waveform in (seq_len, 1)
            x_len                 (int) - the waveform length :code:`seq_len`
            unique_name           (str) - the unique id for this datapoint

        """
        assert mode in [
            "train",
            "test",
        ], "Only support train & test mode (no validation)"

        if mode == "train":

            @dataclass
            class Config:
                min_secs: float = None
                max_secs: float = None

            conf = build_dataset.get("train", {})
            conf = Config(**conf)

            csv = pd.read_csv(data_csv)
            wav_paths = csv["wav_path"].tolist()
            audio_loader = LoadAudio(
                wav_paths, sox_effects=EFFECTS, max_secs=conf.max_secs
            )

            labels = csv["spk"].tolist()
            with open(encoder_path, "rb") as f:
                encoder = pickle.load(f)

            label_encoder = EncodeCategory(labels, encoder)
            ids = csv["id"].tolist()

            class SVTrainDataset:
                def __len__(self):
                    return len(audio_loader)

                def __getitem__(self, index: int):
                    audio = audio_loader[index]
                    label = label_encoder[index]
                    return {
                        "x": audio["wav"],
                        "x_len": audio["wav_len"],
                        "class_id": label["class_id"],
                        "unique_name": ids[index],
                    }

            dataset = SVTrainDataset()

            if conf.min_secs is not None:
                x_lens, unique_names = get_info(
                    dataset,
                    ["x_len", "unique_name"],
                    target_dir / "train_utt_len",
                )

                indices = []
                removed_indices = []
                for idx, (x_len, unique_name) in enumerate(zip(x_lens, unique_names)):
                    secs = x_len / SAMPLE_RATE
                    if secs <= conf.min_secs:
                        logger.info(
                            f"Remove utt {unique_name} since too short after sox effects: {secs} secs"
                        )
                        removed_indices.append(idx)
                    else:
                        indices.append(idx)

                if len(removed_indices) > 0:
                    logger.info(f"Remove in total {len(removed_indices)} utts")

                dataset = Subset(dataset, indices)

        elif mode == "test":
            csv = pd.read_csv(data_csv)
            ids = pd.concat([csv["id1"], csv["id2"]], ignore_index=True).tolist()
            wav_paths = pd.concat(
                [csv["wav_path1"], csv["wav_path2"]], ignore_index=True
            ).tolist()
            data_list = sorted(set([(idx, path) for idx, path in zip(ids, wav_paths)]))
            ids, wav_paths = zip(*data_list)

            audio_loader = LoadAudio(wav_paths)

            class SVTestDataset:
                def __len__(self):
                    return len(audio_loader)

                def __getitem__(self, index: int):
                    audio = audio_loader[index]
                    return {
                        "x": audio["wav"],
                        "x_len": audio["wav_len"],
                        "unique_name": ids[index],
                    }

            dataset = SVTestDataset()

        return dataset

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset,
    ):
        """
        Return the batch sampler for torch DataLoader.

        Args:
            build_batch_sampler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                train                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                test                  (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                ====================  ====================

                Note that ASV does not support valid

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): the :code:`mode` specific csv from :obj:`prepare_data`
            dataset: the dataset from :obj:`build_dataset`

        Returns:
            batch sampler for torch DataLoader
        """

        train = build_batch_sampler.get("train", {})
        test = build_batch_sampler.get("test", {})

        if mode == "train":
            return FixedBatchSizeBatchSampler(dataset, **train)
        elif mode == "test":
            return FixedBatchSizeBatchSampler(dataset, **test)
        else:
            raise ValueError("ASV only supports train/test modes")

    def build_downstream(
        self,
        build_downstream: dict,
        downstream_input_size: int,
        downstream_output_size: int,
        downstream_input_stride: int,
    ):
        """
        Return the task-specific downstream model.
        By default build the :obj:`SuperbXvector` model

        Args:
            build_downstream (dict): same in :obj:`default_config`, support arguments of :obj:`SuperbXvector`
            downstream_input_size (int): the required input size of the model
            downstream_output_size (int): the required output size of the model
            downstream_input_stride (int): the input feature's stride (from 16 KHz)

        Returns:
            :obj:`s3prl.nn.interface.AbsUtteranceModel`
        """
        model = SuperbXvector(downstream_input_size, **build_downstream)
        return model



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# --------------------------------------------------------

import torch
import torch.nn as nn


class SConv1d(nn.Conv1d):
    """SCon1d (Scaling Conv1d): support variable in channels and out channels.

    Notes
    -----
        WeightNorm has `weight_v` and `weight_g`, where
            weight = weight_g / ||weight_v|| * weight_v
        weight is obtained before forward.

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: int = 0,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        """Add dynamic hyper-parameters"""
        super(SConv1d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )
        self.staticize()

    def staticize(self):
        self.sample_in_channels = None
        self.sample_out_channels = None

    def set_sample_config(self, sample_in_channels, sample_out_channels):
        """Sampling subnet"""
        self.sample_in_channels = sample_in_channels
        self.sample_out_channels = sample_out_channels

    def _sample_parameters(self, weight, bias=None):
        """Use to method: set_sample_config"""
        out_dim = self.sample_out_channels
        in_dim = self.sample_in_channels // self.groups
        weight = weight[:out_dim, :in_dim, :]
        if bias is not None:
            bias = bias[:out_dim]
        return weight, bias

    def forward(self, x):
        weight, bias = self._sample_parameters(self.weight, self.bias)
        return self._conv_forward(x, weight, bias)

    def calc_sampled_param_num(self):
        """Calculating the parameters of a subnet"""
        if self.sample_in_channels is not None:
            weight_numel = self.sample_in_channels * self.sample_out_channels
        else:
            weight_numel = self.in_features * self.out_features
        weight_numel = weight_numel * self.kernel_size[0] // self.groups

        if self.bias is not None:
            if self.sample_in_channels is not None:
                bias_numel = self.sample_out_channels
            else:
                bias_numel = self.out_features
        else:
            bias_numel = 0

        return weight_numel + bias_numel

    def get_complexity(self, sequence_length):
        """Getting computational complexity.
        ref: https://github.com/Lyken17/pytorch-OpCounter/blob/509fb7e7e48faaddd4b436371ccd39ede77f1f4a/thop/vision/counter.py#L16
        """
        total_flops = 0
        output_size = sequence_length * self.samples["weight"].size(0)
        in_channel_div_groups = self.samples["weight"].size(1)
        kernel_size = self.kernel_size
        bias = 1 if self.samples["bias"] is not None else 0
        total_flops += output_size * (in_channel_div_groups * kernel_size + bias)
        return total_flops

    def clone_model(self, in_channels: int, out_channels: int):
        """Clone a subnet instance with its supernet's weights."""
        self.set_sample_config(in_channels, out_channels)
        isbias = self.bias is not None
        m = nn.Conv1d(
            in_channels,
            out_channels,
            self.kernel_size,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
            isbias,
            self.padding_mode,
        )
        m = m.to(self.weight.device)
        m = m.to(self.weight.dtype)
        weight, bias = self._sample_parameters(self.weight, self.bias)
        m.weight.data.copy_(weight)
        if isbias:
            m.bias.data.copy_(bias)
        return m.eval()

    @classmethod
    def build_from(cls, m: nn.Conv1d):
        isbias = m.bias is not None
        _m = cls(
            m.in_channels,
            m.out_channels,
            m.kernel_size,
            m.stride,
            m.padding,
            m.dilation,
            m.groups,
            isbias,
            m.padding_mode,
        )
        _m = _m.to(m.weight.device)
        _m = _m.to(m.weight.dtype)
        _m.weight.data.copy_(m.weight)
        if isbias:
            _m.bias.data.copy_(m.bias)
        return _m


if __name__ == "__main__":
    print(SConv1d.__base__)
    m = SConv1d(4, 4, 2, padding=2, groups=2)
    z = m.clone_model(2, 2)
    x = SConv1d.build_from(z)
    x.set_sample_config(2, 2)
    print(m, z, x)
    inp = torch.rand((1, 2, 3))
    m.set_sample_config(2, 2)
    print(torch.allclose(m(inp), z(inp)))
    print(torch.allclose(m(inp), x(inp)))
    m.set_sample_config(2, 1)
    print(type(m.weight))
    print(m.weight.dtype)
    print(m.weight.shape)
    print(m.weight)



"""
Encode the raw data into numeric format, and then decode it
"""

from .category import CategoryEncoder, CategoryEncoders
from .g2p import G2P
from .tokenizer import (
    BertTokenizer,
    CharacterSlotTokenizer,
    CharacterTokenizer,
    SubwordSlotTokenizer,
    SubwordTokenizer,
    Tokenizer,
    WordTokenizer,
)
from .vocabulary import generate_basic_vocab, generate_subword_vocab, generate_vocab

__all__ = [
    "CategoryEncoder",
    "CategoryEncoders",
    "G2P",
    "Tokenizer",
    "BertTokenizer",
    "WordTokenizer",
    "CharacterTokenizer",
    "CharacterSlotTokenizer",
    "SubwordTokenizer",
    "SubwordSlotTokenizer",
    "generate_basic_vocab",
    "generate_subword_vocab",
    "generate_vocab",
]



"""
Speaker verification models

Authors:
  * Haibin Wu 2022
"""

import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from .pooling import (
    AttentiveStatisticsPooling,
    SelfAttentivePooling,
    TemporalAveragePooling,
    TemporalStatisticsPooling,
)

XVECTOR_TDNNS_LENGTH_REDUCTION = 14
ECAPA_TDNNS_LENGTH_REDUCTION = 0


__all__ = [
    "TDNN",
    "XVectorBackbone",
    "ECAPA_TDNN",
    "SpeakerEmbeddingExtractor",
    "SuperbXvector",
]


class TDNN(nn.Module):
    """
    TDNN as defined by https://www.danielpovey.com/files/2015_interspeech_multisplice.pdf.

    Context size and dilation determine the frames selected
    (although context size is not really defined in the traditional sense).

    For example:

        context size 5 and dilation 1 is equivalent to [-2,-1,0,1,2]

        context size 3 and dilation 2 is equivalent to [-2, 0, 2]

        context size 1 and dilation 1 is equivalent to [0]

    Args:
        input_size (int): The input feature size
        output_size (int): The output feature size
        context_size (int): See example
        dilation (int): See example
        dropout_p (float): (default, 0.0) The dropout rate
        batch_norm (bool): (default, False) Use batch norm for TDNN layers
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        context_size: int,
        dilation: int,
        dropout_p: float = 0.0,
        batch_norm: bool = True,
    ):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size
        self.context_size = context_size
        self.dilation = dilation
        self.dropout_p = dropout_p
        self.batch_norm = batch_norm

        self.kernel = nn.Linear(input_size * context_size, output_size)
        self.nonlinearity = nn.ReLU()
        if batch_norm:
            self.bn = nn.BatchNorm1d(output_size)
        if dropout_p:
            self.drop = nn.Dropout(p=dropout_p)

    @property
    def input_size(self) -> int:
        return self._indim

    @property
    def output_size(self) -> int:
        return self._outdim

    def forward(self, x: torch.Tensor):
        """
        Args:
            x (torch.FloatTensor): (batch, seq_len, input_size)

        Returns:
            torch.FloatTensor: (batch, seq_len, output_size)
        """

        _, _, d = x.shape
        assert (
            d == self.input_size
        ), "Input size was wrong. Expected ({}), got ({})".format(self.input_size, d)
        x = x.unsqueeze(1)

        # Unfold input into smaller temporal contexts
        x = F.unfold(
            x,
            (self.context_size, self.input_size),
            stride=(1, self.input_size),
            dilation=(self.dilation, 1),
        )

        x = x.transpose(1, 2)
        x = self.kernel(x)
        x = self.nonlinearity(x)

        if self.dropout_p:
            x = self.drop(x)

        if self.batch_norm:
            x = x.transpose(1, 2)
            x = self.bn(x)
            x = x.transpose(1, 2)

        return x


class XVectorBackbone(nn.Module):
    """
    The TDNN layers the same as in https://danielpovey.com/files/2018_odyssey_xvector_lid.pdf.

    Args:
        input_size (int): The input feature size, usually is the output size of upstream models
        output_size (int): (default, 1500) The size of the speaker embedding
        dropout_p (float): (default, 0.0) The dropout rate
        batch_norm (bool): (default, False) Use batch norm for TDNN layers
    """

    def __init__(
        self,
        input_size: int,
        output_size: int = 1500,
        dropout_p: float = 0.0,
        batch_norm: False = True,
    ):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size

        self.module = nn.Sequential(
            TDNN(
                input_size=input_size,
                output_size=512,
                context_size=5,
                dilation=1,
                dropout_p=dropout_p,
                batch_norm=batch_norm,
            ),
            TDNN(
                input_size=512,
                output_size=512,
                context_size=3,
                dilation=2,
                dropout_p=dropout_p,
                batch_norm=batch_norm,
            ),
            TDNN(
                input_size=512,
                output_size=512,
                context_size=3,
                dilation=3,
                dropout_p=dropout_p,
                batch_norm=batch_norm,
            ),
            TDNN(
                input_size=512,
                output_size=512,
                context_size=1,
                dilation=1,
                dropout_p=dropout_p,
                batch_norm=batch_norm,
            ),
            TDNN(
                input_size=512,
                output_size=output_size,
                context_size=1,
                dilation=1,
                dropout_p=dropout_p,
                batch_norm=batch_norm,
            ),
        )

    @property
    def input_size(self) -> int:
        return self._indim

    @property
    def output_size(self) -> int:
        return self._outdim

    def forward(self, x: torch.Tensor):
        """
        Args:
            x (torch.FloatTensor): (batch, seq_len, input_size)

        output:
            torch.FloatTensor: (batch, seq_len, output_size)
        """
        x = self.module(x)
        return x


"""
ECAPA-TDNN
"""


class _SEModule(nn.Module):
    def __init__(self, channels, bottleneck=128):
        super().__init__()
        self.se = nn.Sequential(
            nn.AdaptiveAvgPool1d(1),
            nn.Conv1d(channels, bottleneck, kernel_size=1, padding=0),
            nn.ReLU(),
            nn.Conv1d(bottleneck, channels, kernel_size=1, padding=0),
            nn.Sigmoid(),
        )

    def forward(self, input):
        x = self.se(input)
        return input * x


class _Bottle2neck(nn.Module):
    def __init__(self, inplanes, planes, kernel_size=None, dilation=None, scale=8):
        super().__init__()
        width = int(math.floor(planes / scale))
        self.conv1 = nn.Conv1d(inplanes, width * scale, kernel_size=1)
        self.bn1 = nn.BatchNorm1d(width * scale)
        self.nums = scale - 1
        convs = []
        bns = []
        num_pad = math.floor(kernel_size / 2) * dilation
        for i in range(self.nums):
            convs.append(
                nn.Conv1d(
                    width,
                    width,
                    kernel_size=kernel_size,
                    dilation=dilation,
                    padding=num_pad,
                )
            )
            bns.append(nn.BatchNorm1d(width))
        self.convs = nn.ModuleList(convs)
        self.bns = nn.ModuleList(bns)
        self.conv3 = nn.Conv1d(width * scale, planes, kernel_size=1)
        self.bn3 = nn.BatchNorm1d(planes)
        self.relu = nn.ReLU()
        self.width = width
        self.se = _SEModule(planes)

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.relu(out)
        out = self.bn1(out)

        spx = torch.split(out, self.width, 1)
        for i in range(self.nums):
            if i == 0:
                sp = spx[i]
            else:
                sp = sp + spx[i]
            sp = self.convs[i](sp)
            sp = self.relu(sp)
            sp = self.bns[i](sp)
            if i == 0:
                out = sp
            else:
                out = torch.cat((out, sp), 1)
        out = torch.cat((out, spx[self.nums]), 1)

        out = self.conv3(out)
        out = self.relu(out)
        out = self.bn3(out)

        out = self.se(out)
        out += residual
        return out


class ECAPA_TDNN(nn.Module):
    """
    ECAPA-TDNN model as in https://arxiv.org/abs/2005.07143.

    Reference code: https://github.com/TaoRuijie/ECAPA-TDNN.

    Args:
        input_size (int): The input feature size, usually is the output size of upstream models
        output_size (int): (default, 1536) The size of the speaker embedding
        C (int): (default, 1024) The channel dimension
    """

    def __init__(
        self, input_size: int = 80, output_size: int = 1536, C: int = 1024, **kwargs
    ):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size

        self.conv1 = nn.Conv1d(input_size, C, kernel_size=5, stride=1, padding=2)
        self.relu = nn.ReLU()
        self.bn1 = nn.BatchNorm1d(C)
        self.layer1 = _Bottle2neck(C, C, kernel_size=3, dilation=2, scale=8)
        self.layer2 = _Bottle2neck(C, C, kernel_size=3, dilation=3, scale=8)
        self.layer3 = _Bottle2neck(C, C, kernel_size=3, dilation=4, scale=8)
        self.layer4 = nn.Conv1d(3 * C, output_size, kernel_size=1)

    @property
    def input_size(self):
        return self._indim

    @property
    def output_size(self):
        return self._outdim

    def forward(self, x: torch.FloatTensor):
        """
        Args:
            x (torch.FloatTensor): size (batch, seq_len, input_size)

        Returns:
            x (torch.FloatTensor): size (batch, seq_len, output_size)
        """

        x = self.conv1(x.transpose(1, 2).contiguous())
        x = self.relu(x)
        x = self.bn1(x)

        x1 = self.layer1(x)
        x2 = self.layer2(x + x1)
        x3 = self.layer3(x + x1 + x2)

        x = self.layer4(torch.cat((x1, x2, x3), dim=1))
        x = self.relu(x)
        x = x.transpose(1, 2).contiguous()

        return x


class SpeakerEmbeddingExtractor(nn.Module):
    """
    The speaker embedding extractor module.

    Args:
        input_size (int): The input feature size, usually is the output size of upstream models
        output_size (int): (default, 1500) The size of the speaker embedding
        backbone (str): (default, XVector) Use which kind of speaker model
        pooling_type (str): (default, TAP) Use which kind of pooling method
    """

    def __init__(
        self,
        input_size: int,
        output_size: int = 1500,
        backbone: str = "XVector",
        pooling_type: str = "TemporalAveragePooling",
    ):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size

        if backbone == "XVector":
            self.backbone = XVectorBackbone(
                input_size=input_size, output_size=output_size
            )
            self.offset = XVECTOR_TDNNS_LENGTH_REDUCTION

        elif backbone == "ECAPA-TDNN":
            self.backbone = ECAPA_TDNN(input_size=input_size, output_size=output_size)
            self.offset = ECAPA_TDNNS_LENGTH_REDUCTION

        else:
            raise ValueError("{} backbone type is not defined".format(backbone))

        if pooling_type == "TemporalAveragePooling" or pooling_type == "TAP":
            self.pooling = TemporalAveragePooling(self.backbone.output_size)

        elif pooling_type == "TemporalStatisticsPooling" or pooling_type == "TSP":
            self.pooling = TemporalStatisticsPooling(self.backbone.output_size)

        elif pooling_type == "SelfAttentivePooling" or pooling_type == "SAP":
            self.pooling = SelfAttentivePooling(self.backbone.output_size)

        elif pooling_type == "AttentiveStatisticsPooling" or pooling_type == "ASP":
            self.pooling = AttentiveStatisticsPooling(self.backbone.output_size)

        else:
            raise ValueError("{} pooling type is not defined".format(pooling_type))

        self._outdim = self.pooling.output_size

    @property
    def input_size(self) -> int:
        return self._indim

    @property
    def output_size(self) -> int:
        return self._outdim

    def forward(self, x: torch.Tensor, xlen: torch.LongTensor = None):
        """
        Args:
            x (torch.Tensor): size (batch, seq_len, input_size)
            xlen (torch.LongTensor): size (batch, )

        Returns:
            x (torch.Tensor): size (batch, output_size)
        """

        x = self.backbone(x)

        if xlen is not None:
            xlen = torch.LongTensor([max(item - self.offset, 0) for item in xlen])
        else:
            xlen = torch.LongTensor([x.shape[1]] * x.shape[0])

        x = self.pooling(x, xlen)

        return x


class _UtteranceExtractor(nn.Module):
    def __init__(self, input_size, output_size):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size

        self.linear1 = nn.Linear(input_size, output_size)
        self.linear2 = nn.Linear(output_size, output_size)
        self.act_fn = nn.ReLU()

    @property
    def input_size(self):
        return self._indim

    @property
    def output_size(self):
        return self._outdim

    def forward(self, x_BxH):
        hid_BxH = self.linear1(x_BxH)
        hid_BxH = self.act_fn(hid_BxH)

        if self.training:
            hid_BxH = self.linear2(hid_BxH)
            hid_BxH = self.act_fn(hid_BxH)

        return hid_BxH


class SuperbXvector(nn.Module):
    """
    The Xvector used in the SUPERB Benchmark with the exact default arguments.

    Args:
        input_size (int): The input feature size, usually is the output size of upstream models
        output_size (int): (default, 512) The size of the speaker embedding
        hidden_size (int): (default, 512) The major hidden size in the network
        aggregation_size (int): (default, 1500) The output size of the x-vector, which is usually large
        dropout_p (float): (default, 0.0) The dropout rate
        batch_norm (bool): (default, False) Use batch norm for TDNN layers
    """

    def __init__(
        self,
        input_size: int,
        output_size: int = 512,
        hidden_size: int = 512,
        aggregation_size: int = 1500,
        dropout_p: float = 0.0,
        batch_norm: bool = False,
    ):
        super().__init__()
        self._input_size = input_size
        self._output_size = output_size

        self.projector = nn.Linear(input_size, hidden_size)
        self.tdnns = XVectorBackbone(
            hidden_size, aggregation_size, dropout_p=dropout_p, batch_norm=batch_norm
        )
        latest_size = self.tdnns.output_size

        self.pooling = TemporalStatisticsPooling(latest_size)
        latest_size = self.pooling.output_size

        self.affine = _UtteranceExtractor(latest_size, output_size)

    @property
    def input_size(self) -> int:
        return self._input_size

    @property
    def output_size(self) -> int:
        return self._output_size

    def forward(self, x, x_len):
        """
        Args:
            x (torch.FloatTensor): (batch_size, seq_len, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Returns:
            torch.FloatTensor: (batch_size, output_size)
        """

        x = self.projector(x)

        x = self.tdnns(x)
        x_len = x_len - XVECTOR_TDNNS_LENGTH_REDUCTION
        assert (
            x_len <= 0
        ).sum() == 0, "The input sequence is too short for the X-vector model"

        x = self.pooling(x, x_len)
        x = self.affine(x)
        return x



"""
The setting of Superb SF

Authors
  * Yung-Sung Chuang 2021
  * Heng-Jui Chang 2022
  * Leo 2022
"""

import pickle
import re
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import List

import pandas as pd
from omegaconf import MISSING

from s3prl.dataio.corpus.snips import SNIPS
from s3prl.dataio.dataset import EncodeText, LoadAudio, get_info
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler, SortedSliceSampler

from .superb_asr import SuperbASR, prepare_common_tokenizer

# Mapping for character-slot tokenizer (SNIPS)
translator = str.maketrans('ÁÃÄÅÆÇÈÉÊËÍÏÐÒÓÔÖØÚÛĘŃŌŞŪ"', "AAAAACEEEEIIDOOOOOUUENOSU ")

__all__ = [
    "audio_snips_for_slot_filling",
    "SuperbSF",
]


def audio_snips_for_slot_filling(
    target_dir: str,
    cache_dir: str,
    dataset_root: str,
    train_speakers: List[str],
    valid_speakers: List[str],
    test_speakers: List[str],
    get_path_only: bool = False,
):
    target_dir = Path(target_dir)

    train_path = target_dir / f"train.csv"
    valid_path = target_dir / f"valid.csv"
    test_paths = [target_dir / f"test.csv"]

    if get_path_only:
        return train_path, valid_path, test_paths

    corpus = SNIPS(dataset_root, train_speakers, valid_speakers, test_speakers)
    train_data, valid_data, test_data = corpus.data_split

    def dict_to_csv(data_dict, csv_path):
        data_ids = sorted(list(data_dict.keys()))
        fields = sorted(data_dict[data_ids[0]].keys())
        data = defaultdict(list)
        for data_id in data_ids:
            data_point = data_dict[data_id]

            trans = data_point["transcription"]
            trans = trans.replace("楽園追放", "EXPELLED")
            trans = trans.replace("官方杂志", "")
            trans = trans.replace("–", "-")
            trans = trans.replace("&", " AND ")
            trans = trans.translate(translator)
            trans = re.sub(" +", " ", trans).strip(" ")

            words = trans.split(" ")
            iobs = data_point["iob"].split(" ")
            assert len(words) == len(iobs)

            filtered_words = []
            filtered_iobs = []
            for word, iob in zip(words, iobs):
                if word in "?!.,;-–…":
                    continue
                filtered_words.append(word)
                filtered_iobs.append(iob)

            assert len(filtered_words) == len(filtered_iobs)
            data_point["transcription"] = " ".join(filtered_words)
            data_point["iob"] = " ".join(filtered_iobs)

            for field in fields:
                data[field].append(data_point[field])

        data["id"] = data_ids
        df = pd.DataFrame(data)
        df.to_csv(csv_path, index=False)

    dict_to_csv(train_data, train_path)
    dict_to_csv(valid_data, valid_path)
    dict_to_csv(test_data, test_paths[0])

    return train_path, valid_path, test_paths


class SuperbSF(SuperbASR):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
                train_speakers=[
                    "Ivy",
                    "Joanna",
                    "Joey",
                    "Justin",
                    "Kendra",
                    "Kimberly",
                    "Matthew",
                    "Salli",
                ],
                valid_speakers=["Aditi", "Amy", "Geraint", "Nicole"],
                test_speakers=["Brian", "Emma", "Raveena", "Russell"],
            ),
            prepare_tokenizer_data=dict(),
            build_tokenizer=dict(
                vocab_type="character",
            ),
            build_dataset=dict(),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=32,
                    max_length=300000,
                ),
                valid=dict(
                    batch_size=1,
                ),
                test=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                model_conf=dict(
                    module="LSTM",
                    proj_size=1024,
                    hidden_size=[1024, 1024],
                    dropout=[0.2, 0.2],
                    layer_norm=[False, False],
                    proj=[False, False],
                    sample_rate=[1, 1],
                    sample_style="concat",
                    bidirectional=True,
                ),
                specaug_conf=dict(
                    freq_mask_width_range=(0, 50),
                    num_freq_mask=4,
                    time_mask_width_range=(0, 40),
                    num_time_mask=2,
                ),
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                log_metrics=[
                    "wer",
                    "cer",
                    "slot_type_f1",
                    "slot_value_cer",
                    "slot_value_wer",
                    "slot_edit_f1_full",
                    "slot_edit_f1_part",
                ],
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-4,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=200000,
                log_step=100,
                eval_step=2000,
                save_step=500,
                gradient_clipping=1.0,
                gradient_accumulate=1,
                valid_metric="slot_type_f1",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        """
        Prepare the task-specific data metadata (path, labels...).
        By default call :obj:`audio_snips_for_slot_filling` with :code:`**prepare_data`

        Args:
            prepare_data (dict): same in :obj:`default_config`, support arguments in :obj:`audio_snips_for_slot_filling`
            target_dir (str): Parse your corpus and save the csv file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            tuple

            1. train_path (str)
            2. valid_path (str)
            3. test_paths (List[str])

            Each path (str) should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            id                    (str) - the unique id for this data point
            wav_path              (str) - the absolute path of the waveform file
            transcription         (str) - a text string where words are separted by a space.
                                    Eg. "I want to fly from Taipei to New York"
            iob                   (str) - iob tags, use "O" if no tag, every word should have a tag, separted by a space.
                                    Eg. "O O O O O from_location O to_location to_location"
            ====================  ====================
        """
        return audio_snips_for_slot_filling(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )

    def prepare_tokenizer_data(
        self,
        prepare_tokenizer_data: dict,
        target_dir: str,
        cache_dir: str,
        train_csv: str,
        valid_csv: str,
        test_csvs: str,
        get_path_only: bool = False,
    ):
        data_dir = target_dir / "tokenizer_data"
        if get_path_only:
            return data_dir

        train_df = pd.read_csv(train_csv)
        valid_df = pd.read_csv(valid_csv)
        test_dfs = [pd.read_csv(test_csv) for test_csv in test_csvs]
        iob_lines = pd.concat([train_df, valid_df, *test_dfs], axis=0)["iob"].tolist()
        iobs = []
        for line in iob_lines:
            iobs.extend(line.split(" "))
        iobs = list(sorted(set(iobs)))

        Path(data_dir).mkdir(parents=True, exist_ok=True)

        with open(data_dir / "slot.txt", "w") as f:
            f.writelines([f"{iob}\n" for iob in iobs])

        train_df = pd.read_csv(train_csv)
        texts = train_df["transcription"].tolist()

        with open(data_dir / "text.txt", "w") as f:
            f.writelines([f"{t}\n" for t in texts])

        return data_dir

    def build_tokenizer(
        self,
        build_tokenizer: dict,
        target_dir: str,
        cache_dir: str,
        tokenizer_data_path: str,
        get_path_only: bool = False,
    ):
        return prepare_common_tokenizer(
            target_dir,
            cache_dir,
            Path(tokenizer_data_path) / "text.txt",
            get_path_only,
            None,
            None,
            slots_file=Path(tokenizer_data_path) / "slot.txt",
            **build_tokenizer,
        )

    def build_dataset(
        self,
        build_dataset: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        tokenizer_path: str,
    ):
        csv = pd.read_csv(data_csv)

        audio_loader = LoadAudio(csv["wav_path"].tolist())

        with open(tokenizer_path, "rb") as f:
            tokenizer = pickle.load(f)

        text_encoder = EncodeText(
            csv["transcription"].tolist(), tokenizer, iob=csv["iob"].tolist()
        )
        ids = csv["id"].tolist()

        class SlotFillingDataset:
            def __len__(self):
                return len(audio_loader)

            def __getitem__(self, index: int):
                audio = audio_loader[index]
                text = text_encoder[index]
                return {
                    "x": audio["wav"],
                    "x_len": audio["wav_len"],
                    "class_ids": text["class_ids"],
                    "labels": text["labels"],
                    "unique_name": ids[index],
                }

        dataset = SlotFillingDataset()
        return dataset

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset,
    ):
        """
        Return the batch sampler for torch DataLoader.

        Args:
            build_batch_sampler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                train                 (dict) - arguments for :obj:`SortedSliceSampler`
                valid                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                test                  (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                ====================  ====================

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): the :code:`mode` specific csv from :obj:`prepare_data`
            dataset: the dataset from :obj:`build_dataset`

        Returns:
            batch sampler for torch DataLoader
        """

        @dataclass
        class Config:
            train: dict = None
            valid: dict = None
            test: dict = None

        conf = Config(**build_batch_sampler)

        if mode == "train":
            wav_lens = get_info(dataset, ["x_len"], Path(target_dir) / "train_stats")
            sampler = SortedSliceSampler(wav_lens, **(conf.train or {}))
            return sampler
        elif mode == "valid":
            return FixedBatchSizeBatchSampler(dataset, **(conf.valid or {}))
        elif mode == "test":
            return FixedBatchSizeBatchSampler(dataset, **(conf.test or {}))
        else:
            raise ValueError(f"Unsupported mode: {mode}")



"""
Most of this code comes from the timm  library.
We tried to disentangle from the timm library version.

Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py

"""
import collections
import logging
import math
import warnings
from collections import OrderedDict
from copy import deepcopy
from functools import partial
from itertools import repeat

import torch
import torch.nn as nn
import torch.nn.functional as F

from .helpers.vit_helpers import (
    DropPath,
    build_model_with_cfg,
    trunc_normal_,
    update_default_cfg_and_kwargs,
)

_logger = logging.getLogger()

IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)
IMAGENET_INCEPTION_MEAN = (0.5, 0.5, 0.5)
IMAGENET_INCEPTION_STD = (0.5, 0.5, 0.5)


# From PyTorch internals
def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))

    return parse


to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)
to_3tuple = _ntuple(3)
to_4tuple = _ntuple(4)
to_ntuple = _ntuple


def _cfg(url="", **kwargs):
    return {
        "url": url,
        "num_classes": 1000,
        "input_size": (3, 224, 224),
        "pool_size": None,
        "crop_pct": 0.9,
        "interpolation": "bicubic",
        "fixed_input_size": True,
        "mean": IMAGENET_INCEPTION_MEAN,
        "std": IMAGENET_INCEPTION_STD,
        "first_conv": "patch_embed.proj",
        "classifier": "head",
        **kwargs,
    }


default_cfgs = {
    # patch models (weights from official Google JAX impl)
    "vit_tiny_patch16_224": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "Ti_16-i21k-300ep-lr_0.001-aug_none-wd_0.03-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.03-res_224.npz"
    ),
    "vit_tiny_patch16_384": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "Ti_16-i21k-300ep-lr_0.001-aug_none-wd_0.03-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.03-res_384.npz",
        input_size=(3, 384, 384),
        crop_pct=1.0,
    ),
    "vit_small_patch32_224": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "S_32-i21k-300ep-lr_0.001-aug_light1-wd_0.03-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.03-res_224.npz"
    ),
    "vit_small_patch32_384": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "S_32-i21k-300ep-lr_0.001-aug_light1-wd_0.03-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.03-res_384.npz",
        input_size=(3, 384, 384),
        crop_pct=1.0,
    ),
    "vit_small_patch16_224": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "S_16-i21k-300ep-lr_0.001-aug_light1-wd_0.03-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.03-res_224.npz"
    ),
    "vit_small_patch16_384": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "S_16-i21k-300ep-lr_0.001-aug_light1-wd_0.03-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.03-res_384.npz",
        input_size=(3, 384, 384),
        crop_pct=1.0,
    ),
    "vit_base_patch32_224": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "B_32-i21k-300ep-lr_0.001-aug_medium1-wd_0.03-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.03-res_224.npz"
    ),
    "vit_base_patch32_384": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "B_32-i21k-300ep-lr_0.001-aug_light1-wd_0.1-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.03-res_384.npz",
        input_size=(3, 384, 384),
        crop_pct=1.0,
    ),
    "vit_base_patch16_224": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "B_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.01-res_224.npz"
    ),
    "vit_base_patch16_384": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "B_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.0-sd_0.0--imagenet2012-steps_20k-lr_0.01-res_384.npz",
        input_size=(3, 384, 384),
        crop_pct=1.0,
    ),
    "vit_large_patch32_224": _cfg(
        url="",  # no official model weights for this combo, only for in21k
    ),
    "vit_large_patch32_384": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vitjx/jx_vit_large_p32_384-9b920ba8.pth",
        input_size=(3, 384, 384),
        crop_pct=1.0,
    ),
    "vit_large_patch16_224": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "L_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.1-sd_0.1--imagenet2012-steps_20k-lr_0.01-res_224.npz"
    ),
    "vit_large_patch16_384": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/"
        "L_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.1-sd_0.1--imagenet2012-steps_20k-lr_0.01-res_384.npz",
        input_size=(3, 384, 384),
        crop_pct=1.0,
    ),
    # patch models, imagenet21k (weights from official Google JAX impl)
    "vit_tiny_patch16_224_in21k": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/Ti_16-i21k-300ep-lr_0.001-aug_none-wd_0.03-do_0.0-sd_0.0.npz",
        num_classes=21843,
    ),
    "vit_small_patch32_224_in21k": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/S_32-i21k-300ep-lr_0.001-aug_light1-wd_0.03-do_0.0-sd_0.0.npz",
        num_classes=21843,
    ),
    "vit_small_patch16_224_in21k": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/S_16-i21k-300ep-lr_0.001-aug_light1-wd_0.03-do_0.0-sd_0.0.npz",
        num_classes=21843,
    ),
    "vit_base_patch32_224_in21k": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/B_32-i21k-300ep-lr_0.001-aug_medium1-wd_0.03-do_0.0-sd_0.0.npz",
        num_classes=21843,
    ),
    "vit_base_patch16_224_in21k": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/B_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.0-sd_0.0.npz",
        num_classes=21843,
    ),
    "vit_large_patch32_224_in21k": _cfg(
        url="https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-vitjx/jx_vit_large_patch32_224_in21k-9046d2e7.pth",
        num_classes=21843,
    ),
    "vit_large_patch16_224_in21k": _cfg(
        url="https://storage.googleapis.com/vit_models/augreg/L_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.1-sd_0.1.npz",
        num_classes=21843,
    ),
    "vit_huge_patch14_224_in21k": _cfg(
        url="https://storage.googleapis.com/vit_models/imagenet21k/ViT-H_14.npz",
        hf_hub="timm/vit_huge_patch14_224_in21k",
        num_classes=21843,
    ),
    # SAM trained models (https://arxiv.org/abs/2106.01548)
    "vit_base_patch32_sam_224": _cfg(
        url="https://storage.googleapis.com/vit_models/sam/ViT-B_32.npz"
    ),
    "vit_base_patch16_sam_224": _cfg(
        url="https://storage.googleapis.com/vit_models/sam/ViT-B_16.npz"
    ),
    # deit models (FB weights)
    "deit_tiny_patch16_224": _cfg(
        url="https://dl.fbaipublicfiles.com/deit/deit_tiny_patch16_224-a1311bcf.pth",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    ),
    "deit_small_patch16_224": _cfg(
        url="https://dl.fbaipublicfiles.com/deit/deit_small_patch16_224-cd65a155.pth",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    ),
    "deit_base_patch16_224": _cfg(
        url="https://dl.fbaipublicfiles.com/deit/deit_base_patch16_224-b5f2ef4d.pth",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
    ),
    "deit_base_patch16_384": _cfg(
        url="https://dl.fbaipublicfiles.com/deit/deit_base_patch16_384-8de9b5d1.pth",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(3, 384, 384),
        crop_pct=1.0,
    ),
    "deit_tiny_distilled_patch16_224": _cfg(
        url="https://dl.fbaipublicfiles.com/deit/deit_tiny_distilled_patch16_224-b40b3cf7.pth",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        classifier=("head", "head_dist"),
    ),
    "deit_small_distilled_patch16_224": _cfg(
        url="https://dl.fbaipublicfiles.com/deit/deit_small_distilled_patch16_224-649709d9.pth",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        classifier=("head", "head_dist"),
    ),
    "deit_base_distilled_patch16_224": _cfg(
        url="https://dl.fbaipublicfiles.com/deit/deit_base_distilled_patch16_224-df68dfff.pth",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        classifier=("head", "head_dist"),
    ),
    "deit_base_distilled_patch16_384": _cfg(
        url="https://dl.fbaipublicfiles.com/deit/deit_base_distilled_patch16_384-d0272ac0.pth",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(3, 384, 384),
        crop_pct=1.0,
        classifier=("head", "head_dist"),
    ),
    # ViT ImageNet-21K-P pretraining by MILL
    "vit_base_patch16_224_miil_in21k": _cfg(
        url="https://miil-public-eu.oss-eu-central-1.aliyuncs.com/model-zoo/ImageNet_21K_P/models/timm/vit_base_patch16_224_in21k_miil.pth",
        mean=(0, 0, 0),
        std=(1, 1, 1),
        crop_pct=0.875,
        interpolation="bilinear",
        num_classes=11221,
    ),
    "vit_base_patch16_224_miil": _cfg(
        url="https://miil-public-eu.oss-eu-central-1.aliyuncs.com/model-zoo/ImageNet_21K_P/models/timm"
        "/vit_base_patch16_224_1k_miil_84_4.pth",
        mean=(0, 0, 0),
        std=(1, 1, 1),
        crop_pct=0.875,
        interpolation="bilinear",
    ),
    # PaSST
    "passt_s_swa_p16_128_ap476": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v0.0.1-audioset/passt-s-f128-p16-s10-ap.476-swa.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 998),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=527,
    ),
    "passt_s_p16_s16_128_ap468": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v0.0.2-audioset/passt-s-f128-p16-s16-ap.468.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 998),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=527,
    ),
    "passt_s_swa_f128_stfthop100_p16_s10_ap473": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v0.0.3-audioset/passt-s-f128-stfthop100-p16-s10-ap.473-swa.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 3200),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=527,
    ),
    "passt_s_swa_f128_stfthop160_p16_s10_ap473": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v0.0.3-audioset/passt-s-f128-stfthop160-p16-s10-ap.473-swa.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 2000),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=527,
    ),
    "openmic2008_passt_u_f128_p16_s10_ap85_swa": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v0.0.4-openmic/openmic2008.passt-u-f128-p16-s10-ap.85-swa.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 3200),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=20,
    ),
    "openmic2008_passt_u_f128_p16_s10_ap85  ": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v0.0.4-openmic/openmic2008.passt-u-f128-p16-s10-ap.85.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 2000),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=20,
    ),
    "passt-s-f128-20sec-p16-s10-ap474-swa": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v0.0.5/passt-s-f128-20sec-p16-s10-ap.474-swa.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 2000),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=527,
    ),
    "passt-s-f128-30sec-p16-s10-ap473-swa": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v0.0.5/passt-s-f128-30sec-p16-s10-ap.473-swa.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 3000),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=527,
    ),
    "passt_b_f128_p16_s16_ap_459": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v.0.0.7-audioset/passt-b-f128-p16-s16-ap.459.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 998),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=527,
    ),
    "passt_u600_f128_p16_s16_ap_460": _cfg(
        url="https://github.com/kkoutini/PaSST/releases/download/v.0.0.7-audioset/passt-u600-f128-p16-s16-ap.460.pt",
        mean=IMAGENET_DEFAULT_MEAN,
        std=IMAGENET_DEFAULT_STD,
        input_size=(1, 128, 998),
        crop_pct=1.0,
        classifier=("head.1", "head_dist"),
        num_classes=527,
    ),
}


def adapt_input_conv(in_chans, conv_weight):
    conv_type = conv_weight.dtype
    conv_weight = (
        conv_weight.float()
    )  # Some weights are in torch.half, ensure it's float for sum on CPU
    O, I, J, K = conv_weight.shape
    if in_chans == 1:
        if I > 3:
            assert conv_weight.shape[1] % 3 == 0
            # For models with space2depth stems
            conv_weight = conv_weight.reshape(O, I // 3, 3, J, K)
            conv_weight = conv_weight.sum(dim=2, keepdim=False)
        else:
            conv_weight = conv_weight.sum(dim=1, keepdim=True)
    elif in_chans != 3:
        if I != 3:
            raise NotImplementedError("Weight format not supported by conversion.")
        else:
            # NOTE this strategy should be better than random init, but there could be other combinations of
            # the original RGB input layer weights that'd work better for specific cases.
            repeat = int(math.ceil(in_chans / 3))
            conv_weight = conv_weight.repeat(1, repeat, 1, 1)[:, :in_chans, :, :]
            conv_weight *= 3 / float(in_chans)
    conv_weight = conv_weight.to(conv_type)
    return conv_weight


class Mlp(nn.Module):
    """MLP as used in Vision Transformer, MLP-Mixer and related networks"""

    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class PatchEmbed(nn.Module):
    """2D Image to Patch Embedding"""

    def __init__(
        self,
        img_size=224,
        patch_size=16,
        stride=16,
        in_chans=3,
        embed_dim=768,
        norm_layer=None,
        flatten=True,
    ):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        stride = to_2tuple(stride)
        self.img_size = img_size
        self.patch_size = patch_size
        self.stride = stride
        self.grid_size = (img_size[0] // stride[0], img_size[1] // stride[1])
        self.num_patches = self.grid_size[0] * self.grid_size[1]
        self.flatten = flatten
        self.embed_dim = embed_dim
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=stride
        )
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        B, C, H, W = x.shape
        # if not (H == self.img_size[0] and W == self.img_size[1]):
        #     warnings.warn(f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]}).")
        # to do maybe replace weights
        x = self.proj(x)
        if self.flatten:
            x = x.flatten(2).transpose(1, 2)  # BCHW -> BNC
        x = self.norm(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0.0, proj_drop=0.0):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = (
            self.qkv(x)
            .reshape(B, N, 3, self.num_heads, C // self.num_heads)
            .permute(2, 0, 3, 1, 4)
        )
        q, k, v = (
            qkv[0],
            qkv[1],
            qkv[2],
        )  # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class PaSST(nn.Module):
    """

    Based on the implementation of Vision Transformer in timm library.
     Take a look at the get_model function, adapting the weights of pretrained imagenet models.

    """

    def __init__(
        self,
        u_patchout=0,
        s_patchout_t=0,
        s_patchout_f=0,
        img_size=(128, 998),
        patch_size=16,
        stride=16,
        in_chans=1,
        num_classes=527,
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4.0,
        qkv_bias=True,
        representation_size=None,
        distilled=False,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        embed_layer=PatchEmbed,
        norm_layer=None,
        act_layer=None,
        weight_init="",
    ):
        """
        Args:
            u_patchout: Unstructured Patchout integer, number of items to be removed from the final sequence
            s_patchout_t: structured Patchout time integer, number of columns to be removed from the patches grid
            s_patchout_f: structured Patchout Frequency integer, number of rows to be removed from the patches grid
            img_size (int, tuple): input image size
            patch_size (int, tuple): patch size
            in_chans (int): number of input channels
            num_classes (int): number of classes for classification head
            embed_dim (int): embedding dimension
            depth (int): depth of transformer
            num_heads (int): number of attention heads
            mlp_ratio (int): ratio of mlp hidden dim to embedding dim
            qkv_bias (bool): enable bias for qkv if True
            representation_size (Optional[int]): enable and set representation layer (pre-logits) to this value if set
            distilled (bool): model includes a distillation token and head as in DeiT models
            drop_rate (float): dropout rate
            attn_drop_rate (float): attention dropout rate
            drop_path_rate (float): stochastic depth rate
            embed_layer (nn.Module): patch embedding layer
            norm_layer: (nn.Module): normalization layer
            weight_init: (str): weight init scheme
        """
        super().__init__()
        self.num_classes = num_classes
        self.u_patchout = u_patchout
        self.s_patchout_t = s_patchout_t
        self.s_patchout_f = s_patchout_f
        self.num_features = (
            self.embed_dim
        ) = embed_dim  # num_features for consistency with other models
        self.num_tokens = 2 if distilled else 1
        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        act_layer = act_layer or nn.GELU

        self.patch_embed = embed_layer(
            img_size=img_size,
            patch_size=patch_size,
            stride=stride,
            in_chans=in_chans,
            embed_dim=embed_dim,
            flatten=False,
        )
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.dist_token = (
            nn.Parameter(torch.zeros(1, 1, embed_dim)) if distilled else None
        )
        # PaSST
        # refer to https://arxiv.org/abs/2110.05069 Section 2
        self.new_pos_embed = nn.Parameter(
            torch.zeros(1, self.num_tokens, embed_dim)
        )  # for C and D tokens
        self.freq_new_pos_embed = nn.Parameter(
            torch.zeros(1, embed_dim, self.patch_embed.grid_size[0], 1)
        )  # | f
        self.time_new_pos_embed = nn.Parameter(
            torch.zeros(1, embed_dim, 1, self.patch_embed.grid_size[1])
        )  # __ t
        ####
        self.pos_drop = nn.Dropout(p=drop_rate)

        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, depth)
        ]  # stochastic depth decay rule
        self.blocks = nn.Sequential(
            *[
                Block(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[i],
                    norm_layer=norm_layer,
                    act_layer=act_layer,
                )
                for i in range(depth)
            ]
        )
        self.norm = norm_layer(embed_dim)

        # Representation layer
        if representation_size and not distilled:
            self.num_features = representation_size
            self.pre_logits = nn.Sequential(
                OrderedDict(
                    [
                        ("fc", nn.Linear(embed_dim, representation_size)),
                        ("act", nn.Tanh()),
                    ]
                )
            )
        else:
            self.pre_logits = nn.Identity()

        # Classifier head(s)
        self.head = nn.Sequential(
            nn.LayerNorm(self.num_features),
            nn.Linear(self.num_features, num_classes)
            if num_classes > 0
            else nn.Identity(),
        )
        self.head_dist = None
        if distilled:
            self.head_dist = (
                nn.Linear(self.embed_dim, self.num_classes)
                if num_classes > 0
                else nn.Identity()
            )

        self.init_weights(weight_init)

    def init_weights(self, mode=""):
        assert mode in ("jax", "jax_nlhb", "nlhb", "")
        head_bias = -math.log(self.num_classes) if "nlhb" in mode else 0.0
        trunc_normal_(self.new_pos_embed, std=0.02)
        trunc_normal_(self.freq_new_pos_embed, std=0.02)
        trunc_normal_(self.time_new_pos_embed, std=0.02)
        if self.dist_token is not None:
            trunc_normal_(self.dist_token, std=0.02)
        if mode.startswith("jax"):
            # leave cls token as zeros to match jax impl
            raise RuntimeError("Not supported yet")
        else:
            trunc_normal_(self.cls_token, std=0.02)
            self.apply(_init_vit_weights)

    def _init_weights(self, m):
        # this fn left here for compat with downstream users
        _init_vit_weights(m)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {
            "new_pos_embed",
            "freq_new_pos_embed",
            "time_new_pos_embed",
            "cls_token",
            "dist_token",
        }

    def get_classifier(self):
        if self.dist_token is None:
            return self.head
        else:
            return self.head, self.head_dist

    def reset_classifier(self, num_classes, global_pool=""):
        self.num_classes = num_classes
        self.head = (
            nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()
        )
        if self.num_tokens == 2:
            self.head_dist = (
                nn.Linear(self.embed_dim, self.num_classes)
                if num_classes > 0
                else nn.Identity()
            )

    def forward_features(self, x):
        x = self.patch_embed(x)  # [b, e, f, t]
        B_dim, E_dim, F_dim, T_dim = x.shape  # slow
        # Adding Time/Freq information
        time_new_pos_embed = self.time_new_pos_embed
        if x.shape[-1] != time_new_pos_embed.shape[-1]:
            time_new_pos_embed = time_new_pos_embed[:, :, :, : x.shape[-1]]
        x = x + time_new_pos_embed
        x = x + self.freq_new_pos_embed

        # Structured Patchout https://arxiv.org/abs/2110.05069 Section 2.2
        if self.training and self.s_patchout_t:
            # ([1, 768, 1, 82])
            random_indices = (
                torch.randperm(T_dim)[: T_dim - self.s_patchout_t].sort().values
            )
            x = x[:, :, :, random_indices]
        if self.training and self.s_patchout_f:
            # [1, 768, 12, 1]
            random_indices = (
                torch.randperm(F_dim)[: F_dim - self.s_patchout_f].sort().values
            )
            x = x[:, :, random_indices, :]

        ###
        # Flatten the sequence
        x = x.flatten(2).transpose(1, 2)
        # Unstructured Patchout

        if self.training and self.u_patchout:
            seq_len = x.shape[1]
            random_indices = (
                torch.randperm(seq_len)[: seq_len - self.u_patchout].sort().values
            )
            x = x[:, random_indices, :]

        ####
        # Add the C/D tokens
        cls_tokens = self.cls_token.expand(B_dim, -1, -1) + self.new_pos_embed[:, :1, :]
        if self.dist_token is None:
            x = torch.cat((cls_tokens, x), dim=1)
        else:
            dist_token = (
                self.dist_token.expand(B_dim, -1, -1) + self.new_pos_embed[:, 1:, :]
            )
            x = torch.cat((cls_tokens, dist_token, x), dim=1)

        x = self.pos_drop(x)
        x = self.blocks(x)
        x = self.norm(x)
        if self.dist_token is None:
            return self.pre_logits(x[:, 0])
        else:
            return x[:, 0], x[:, 1]

    def forward(self, x):
        x = self.forward_features(x)

        if self.head_dist is not None:
            features = (x[0] + x[1]) / 2
            x = self.head(features)
            return x, features
        else:
            features = x
            x = self.head(x)

        return x, features


lecun_normal_ = None


def _init_vit_weights(
    module: nn.Module, name: str = "", head_bias: float = 0.0, jax_impl: bool = False
):
    """ViT weight initialization
    * When called without n, head_bias, jax_impl args it will behave exactly the same
      as my original init for compatibility with prev hparam / downstream use cases (ie DeiT).
    * When called w/ valid n (module name) and jax_impl=True, will (hopefully) match JAX impl
    """
    if isinstance(module, nn.Linear):
        if name.startswith("head"):
            nn.init.zeros_(module.weight)
            nn.init.constant_(module.bias, head_bias)
        elif name.startswith("pre_logits"):
            lecun_normal_(module.weight)
            nn.init.zeros_(module.bias)
        else:
            if jax_impl:
                nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    if "mlp" in name:
                        nn.init.normal_(module.bias, std=1e-6)
                    else:
                        nn.init.zeros_(module.bias)
            else:
                trunc_normal_(module.weight, std=0.02)
                if module.bias is not None:
                    nn.init.zeros_(module.bias)
    elif jax_impl and isinstance(module, nn.Conv2d):
        # NOTE conv was left to pytorch default in my original init
        lecun_normal_(module.weight)
        if module.bias is not None:
            nn.init.zeros_(module.bias)
    elif isinstance(module, (nn.LayerNorm, nn.GroupNorm, nn.BatchNorm2d)):
        nn.init.zeros_(module.bias)
        nn.init.ones_(module.weight)


def resize_pos_embed(posemb, posemb_new, num_tokens=1, gs_new=(), mode="bicubic"):
    # Rescale the grid of position embeddings when loading from state_dict. Adapted from
    # https://github.com/google-research/vision_transformer/blob/00883dd691c63a6830751563748663526e811cee/vit_jax/checkpoint.py#L224
    _logger.info(
        "Resized position embedding: %s to %s with %s cls/dis tokens",
        posemb.shape,
        posemb_new.shape,
        num_tokens,
    )
    ntok_new = posemb_new.shape[1]
    if num_tokens:
        posemb_tok, posemb_grid = posemb[:, :num_tokens], posemb[0, num_tokens:]
        ntok_new -= num_tokens
    else:
        posemb_tok, posemb_grid = posemb[:, :0], posemb[0]
    gs_old = int(math.sqrt(len(posemb_grid)))
    if not len(gs_new):  # backwards compatibility
        gs_new = [int(math.sqrt(ntok_new))] * 2
    assert len(gs_new) >= 2
    _logger.info("Position embedding grid-size from %s to %s", [gs_old, gs_old], gs_new)
    posemb_grid = posemb_grid.reshape(1, gs_old, gs_old, -1).permute(0, 3, 1, 2)
    posemb_grid = F.interpolate(
        posemb_grid, size=gs_new, mode=mode, align_corners=False
    )
    posemb_grid = posemb_grid.permute(0, 2, 3, 1).reshape(1, gs_new[0] * gs_new[1], -1)
    posemb = torch.cat([posemb_tok, posemb_grid], dim=1)
    return posemb


def adapt_image_pos_embed_to_passt(posemb, num_tokens=1, gs_new=(), mode="bicubic"):
    # Rescale the grid of position embeddings when loading from state_dict. Adapted from
    # https://github.com/google-research/vision_transformer/blob/00883dd691c63a6830751563748663526e811cee/vit_jax/checkpoint.py#L224
    _logger.info(
        "Resized position embedding: %s to %s with %s cls/dis tokens",
        posemb.shape,
        gs_new,
        num_tokens,
    )
    if num_tokens:
        posemb_tok, posemb_grid = posemb[:, :num_tokens], posemb[0, num_tokens:]
    else:
        posemb_tok, posemb_grid = posemb[:, :0], posemb[0]
    gs_old = int(math.sqrt(len(posemb_grid)))

    assert len(gs_new) >= 2
    _logger.info("Position embedding grid-size from %s to %s", [gs_old, gs_old], gs_new)
    posemb_grid = posemb_grid.reshape(1, gs_old, gs_old, -1).permute(0, 3, 1, 2)
    posemb_grid = F.interpolate(
        posemb_grid, size=gs_new, mode=mode, align_corners=False
    )
    freq_new_pos_embed = posemb_grid.mean(dim=3, keepdim=True)
    time_new_pos_embed = posemb_grid.mean(dim=2, keepdim=True)
    _logger.info("New Position cls/dstl embedding %s", posemb_tok.shape)
    _logger.info("New FREQ Position embedding %s", freq_new_pos_embed.shape)
    _logger.info("New TIME Position embedding %s", time_new_pos_embed.shape)
    return posemb_tok, freq_new_pos_embed, time_new_pos_embed


def checkpoint_filter_fn(state_dict, model):
    """convert patch embedding weight from manual patchify + linear proj to conv"""
    out_dict = {}
    if "model" in state_dict:
        # For deit models
        state_dict = state_dict["model"]
    state_dict = {k: v for k, v in state_dict.items()}
    if "time_new_pos_embed" not in state_dict:
        # we are working with ImageNet model
        _logger.info("Adapting pos embedding from ImageNet pretrained model to PaSST.")
        v = state_dict.pop("pos_embed")
        (
            new_pos_embed,
            freq_new_pos_embed,
            time_new_pos_embed,
        ) = adapt_image_pos_embed_to_passt(
            v, getattr(model, "num_tokens", 1), model.patch_embed.grid_size
        )
        state_dict["new_pos_embed"] = new_pos_embed
        state_dict["freq_new_pos_embed"] = freq_new_pos_embed
        state_dict["time_new_pos_embed"] = time_new_pos_embed

    for k, v in state_dict.items():
        if "patch_embed.proj.weight" in k and len(v.shape) < 4:
            # For old models that I trained prior to conv based patchification
            O, I, H, W = model.patch_embed.proj.weight.shape
            v = v.reshape(O, -1, H, W)
        elif k == "pos_embed" and v.shape != model.pos_embed.shape:
            # this should never occur
            v = resize_pos_embed(
                v,
                model.pos_embed,
                getattr(model, "num_tokens", 1),
                model.patch_embed.grid_size,
            )
        out_dict[k] = v
    return out_dict


def _create_vision_transformer(variant, pretrained=False, default_cfg=None, **kwargs):
    default_cfg = default_cfg or default_cfgs[variant]
    if kwargs.get("features_only", None):
        raise RuntimeError(
            "features_only not implemented for Vision Transformer models."
        )

    # NOTE this extra code to support handling of repr size for in21k pretrained models
    default_num_classes = default_cfg["num_classes"]
    num_classes = kwargs.get("num_classes", default_num_classes)
    repr_size = kwargs.pop("representation_size", None)
    if repr_size is not None and num_classes != default_num_classes:
        # Remove representation layer if fine-tuning. This may not always be the desired action,
        # but I feel better than doing nothing by default for fine-tuning. Perhaps a better interface?
        _logger.warning("Removing representation layer for fine-tuning.")
        repr_size = None

    model = build_model_with_cfg(
        PaSST,
        variant,
        pretrained,
        default_cfg=default_cfg,
        representation_size=repr_size,
        pretrained_filter_fn=checkpoint_filter_fn,
        pretrained_custom_load="npz" in default_cfg["url"],
        **kwargs,
    )
    return model


def vit_huge_patch14_224_in21k(pretrained=False, **kwargs):
    """ViT-Huge model (ViT-H/14) from original paper (https://arxiv.org/abs/2010.11929).
    ImageNet-21k weights @ 224x224, source https://github.com/google-research/vision_transformer.
    NOTE: this model has a representation layer but the 21k classifier head is zero'd out in original weights
    """
    model_kwargs = dict(
        patch_size=14,
        embed_dim=1280,
        depth=32,
        num_heads=16,
        representation_size=1280,
        **kwargs,
    )
    model = _create_vision_transformer(
        "vit_huge_patch14_224_in21k", pretrained=pretrained, **model_kwargs
    )
    return model


def deit_base_distilled_patch16_384(pretrained=False, **kwargs):
    """DeiT-base distilled model @ 384x384 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "deit_base_distilled_patch16_384",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def passt_s_swa_p16_128_ap476(pretrained=False, **kwargs):
    """DeiT-base distilled model @ 384x384 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "passt_s_swa_p16_128_ap476",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def passt_s_p16_s16_128_ap468(pretrained=False, **kwargs):
    """PaSST pre-trained on AudioSet"""
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    if model_kwargs.get("stride") != (16, 16):
        warnings.warn(
            f"This model was pre-trained with strides {(16, 16)}, but now you set (fstride,tstride) to {model_kwargs.get('stride')}."
        )
    model = _create_vision_transformer(
        "passt_s_p16_s16_128_ap468",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def passt_s_swa_f128_stfthop100_p16_s10_ap473(pretrained=False, **kwargs):
    """DeiT-base distilled model @ 384x384 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "passt_s_swa_f128_stfthop100_p16_s10_ap473",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def passt_s_swa_f128_stfthop160_p16_s10_ap473(pretrained=False, **kwargs):
    """DeiT-base distilled model @ 384x384 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "passt_s_swa_f128_stfthop160_p16_s10_ap473",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def openmic2008_passt_u_f128_p16_s10_ap85_swa(pretrained=False, **kwargs):
    """DeiT-base distilled model @ 384x384 from paper (https://arxiv.org/abs/2012.12877).
    ImageNet-1k weights from https://github.com/facebookresearch/deit.
    """
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "openmic2008_passt_u_f128_p16_s10_ap85_swa",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def passt_s_f128_20sec_p16_s10_ap474_swa(pretrained=False, **kwargs):
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "passt-s-f128-20sec-p16-s10-ap474-swa",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def passt_s_f128_30sec_p16_s10_ap473_swa(pretrained=False, **kwargs):
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "passt-s-f128-30sec-p16-s10-ap473-swa",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def passt_b_f128_p16_s16_ap_459(pretrained=False, **kwargs):
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "passt_b_f128_p16_s16_ap_459",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


def passt_u600_f128_p16_s16_ap_460(pretrained=False, **kwargs):
    model_kwargs = dict(patch_size=16, embed_dim=768, depth=12, num_heads=12, **kwargs)
    model = _create_vision_transformer(
        "passt_u600_f128_p16_s16_ap_460",
        pretrained=pretrained,
        distilled=True,
        **model_kwargs,
    )
    return model


PatchEmbedAdaptiveMean = None
PatchEmbedAdaptiveMeanKeepConv = None


def fix_embedding_layer(model, embed="default"):
    if embed == "default":
        return model
    if embed == "overlap":
        model.patch_embed = PatchEmbedAdaptiveMean(replace=model.patch_embed)
    if embed == "am_keepconv":
        model.patch_embed = PatchEmbedAdaptiveMeanKeepConv(replace=model.patch_embed)
    return model


def get_model(
    arch="passt_s_swa_p16_128_ap476",
    pretrained=True,
    n_classes=527,
    in_channels=1,
    fstride=10,
    tstride=10,
    input_fdim=128,
    input_tdim=998,
    u_patchout=0,
    s_patchout_t=0,
    s_patchout_f=0,
):
    """
    :param arch: Base ViT or Deit architecture
    :param pretrained: use pretrained model on imagenet
    :param n_classes: number of classes
    :param in_channels: number of input channels: 1 for mono
    :param fstride: the patches stride over frequency.
    :param tstride: the patches stride over time.
    :param input_fdim: the expected input frequency bins.
    :param input_tdim: the expected input time bins.
    :param u_patchout: number of input patches to drop in Unstructured Patchout as defined in https://arxiv.org/abs/2110.05069
    :param s_patchout_t: number of input time frames to drop Structured Patchout as defined in https://arxiv.org/abs/2110.05069
    :param s_patchout_f:  number of input frequency bins to drop Structured Patchout as defined in https://arxiv.org/abs/2110.05069
    :param audioset_pretrain: use pretrained models on Audioset.
    :return:

    """
    model_func = None
    input_size = (input_fdim, input_tdim)
    stride = (fstride, tstride)
    if arch == "passt_deit_bd_p16_384":
        model_func = deit_base_distilled_patch16_384
    elif arch == "passt_s_swa_p16_128_ap476":  # pretrained
        model_func = passt_s_swa_p16_128_ap476
    elif arch == "passt_s_p16_s16_128_ap468":
        model_func = passt_s_p16_s16_128_ap468
    elif arch == "openmic2008":  # pretrained
        model_func = openmic2008_passt_u_f128_p16_s10_ap85_swa
    elif arch == "stfthop100":  # pretrained
        model_func = passt_s_swa_f128_stfthop100_p16_s10_ap473
    elif arch == "stfthop160":  # pretrained
        model_func = passt_s_swa_f128_stfthop160_p16_s10_ap473
    elif arch == "passt_20sec":  # pretrained
        model_func = passt_s_f128_20sec_p16_s10_ap474_swa
    elif arch == "passt_30sec":  # pretrained
        model_func = passt_s_f128_30sec_p16_s10_ap473_swa
    elif arch == "passt_u600_f128_p16_s16_ap_460":  # pretrained
        model_func = passt_u600_f128_p16_s16_ap_460
    elif arch == "passt_b_f128_p16_s16_ap_459":  # pretrained
        model_func = passt_b_f128_p16_s16_ap_459

    if model_func is None:
        raise RuntimeError(f"Unknown model {arch}")
    model = model_func(
        pretrained=pretrained,
        num_classes=n_classes,
        in_chans=in_channels,
        img_size=input_size,
        stride=stride,
        u_patchout=u_patchout,
        s_patchout_t=s_patchout_t,
        s_patchout_f=s_patchout_f,
    )
    model = fix_embedding_layer(model)
    return model



"""
For datasets with highly unbalanced class

Authors:
  * Leo 2022
"""

from collections import Counter
from typing import Iterator, List, TypeVar

import torch
from torch.utils.data import WeightedRandomSampler

T_co = TypeVar("T_co", covariant=True)

__all__ = ["BalancedWeightedSampler"]


class BalancedWeightedSampler:
    """
    This batch sampler is always randomized, hence cannot be used for testing
    """

    def __init__(
        self,
        labels: List[str],
        batch_size: int,
        duplicate: int = 1,
        seed: int = 12345678,
    ) -> None:
        self.epoch = 0
        self.seed = seed
        self.batch_size = batch_size
        self.duplicate = duplicate

        class2weight = Counter()
        for label in labels:
            class2weight.update([label])

        weights = []
        for label in labels:
            weights.append(len(labels) / class2weight[label])

        self.weights = weights

    def set_epoch(self, epoch: int):
        self.epoch = epoch

    def __iter__(self) -> Iterator[T_co]:
        generator = torch.Generator()
        generator.manual_seed(self.epoch + self.seed)

        sampler = WeightedRandomSampler(
            self.weights, len(self.weights) * self.duplicate, generator=generator
        )
        indices = list(sampler)

        batch = []
        for indice in indices:
            batch.append(indice)
            if len(batch) == self.batch_size:
                yield batch
                batch = []

        if len(batch) > 0:
            yield batch

    def __len__(self):
        return len(list(iter(self)))



from typing import List

import torch

from s3prl.dataio.encoder.category import CategoryEncoder, CategoryEncoders
from s3prl.dataio.encoder.tokenizer import Tokenizer

from . import Dataset

__all__ = [
    "EncodeCategory",
    "EncodeCategories",
    "EncodeMultiLabel",
    "EncodeText",
]


class EncodeCategory(Dataset):
    def __init__(self, labels: List[str], encoder: CategoryEncoder) -> None:
        super().__init__()
        self.labels = labels
        self.encoder = encoder

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, index: int):
        label = self.labels[index]
        return {
            "label": label,
            "class_id": self.encoder.encode(label),
        }


class EncodeCategories(Dataset):
    def __init__(self, labels: List[List[str]], encoders: CategoryEncoders) -> None:
        super().__init__()
        self.labels = labels
        self.encoders = encoders

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, index: int):
        labels = self.labels[index]
        return {
            "labels": labels,
            "class_ids": torch.LongTensor(self.encoders.encode(labels)),
        }


class EncodeMultiLabel(Dataset):
    def __init__(self, labels: List[List[str]], encoder: CategoryEncoder) -> None:
        super().__init__()
        self.labels = labels
        self.encoder = encoder

    def __len__(self):
        return len(self.labels)

    @staticmethod
    def label_to_binary_vector(label_ids: List[int], num_labels: int) -> torch.Tensor:
        if len(label_ids) == 0:
            binary_labels = torch.zeros((num_labels,), dtype=torch.float)
        else:
            binary_labels = torch.zeros((num_labels,)).scatter(
                0, torch.tensor(label_ids), 1.0
            )

        assert set(torch.where(binary_labels == 1.0)[0].numpy()) == set(label_ids)
        return binary_labels

    def __getitem__(self, index: int):
        labels = self.labels[index]
        label_ids = [self.encoder.encode(label) for label in labels]
        binary_labels = self.label_to_binary_vector(label_ids, len(self.encoder))

        return {
            "labels": labels,
            "binary_labels": binary_labels,
        }


class EncodeText(Dataset):
    def __init__(
        self, text: List[str], tokenizer: Tokenizer, iob: List[str] = None
    ) -> None:
        super().__init__()
        self.text = text
        self.iob = iob
        if iob is not None:
            assert len(text) == len(iob)
        self.tokenizer = tokenizer

    def __len__(self):
        return len(self.text)

    def __getitem__(self, index: int):
        text = self.text[index]
        if self.iob is not None:
            iob = self.iob[index]
            tokenized_ids = self.tokenizer.encode(text, iob)
            text = self.tokenizer.decode(tokenized_ids)
        else:
            tokenized_ids = self.tokenizer.encode(text)

        return {
            "labels": text,
            "class_ids": torch.LongTensor(tokenized_ids),
        }



"""
The setting of Superb ASR

Authors
  * Heng-Jui Chang 2022
  * Leo 2022
"""

import logging
import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import List

import pandas as pd
from omegaconf import MISSING
from torch.utils.data import Dataset

from s3prl.dataio.corpus.librispeech import LibriSpeech
from s3prl.dataio.dataset import EncodeText, LoadAudio, get_info
from s3prl.dataio.encoder.tokenizer import load_tokenizer
from s3prl.dataio.encoder.vocabulary import generate_vocab
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler, SortedBucketingSampler
from s3prl.nn.rnn import RNNEncoder
from s3prl.nn.specaug import ModelWithSpecaug
from s3prl.util.download import urls_to_filepaths

from .run import ASR

logger = logging.getLogger(__name__)

__all__ = [
    "prepare_librispeech",
    "prepare_common_tokenizer",
    "SuperbASR",
]


def prepare_librispeech(
    target_dir,
    cache_dir,
    dataset_root,
    train_sets: List[str],
    valid_sets: List[str],
    test_sets: List[str],
    n_jobs: int = 6,
    get_path_only: bool = False,
):
    """
    Prepare LibriSpeech for ASR following :obj:`SuperbASR.prepare_data` format.
    See :obj:`LibriSpeech` for the arguments usage
    """
    target_dir = Path(target_dir)

    train_path = target_dir / f"{'+'.join(train_sets)}.csv"
    valid_path = target_dir / f"{'+'.join(valid_sets)}.csv"
    test_paths = [target_dir / f"{test_set}.csv" for test_set in test_sets]

    if get_path_only:
        return train_path, valid_path, test_paths

    corpus = LibriSpeech(dataset_root, n_jobs, train_sets, valid_sets, test_sets)
    train_data, valid_data, test_data = corpus.data_split

    def dict_to_csv(data_dict, csv_path):
        keys = sorted(list(data_dict.keys()))
        fields = sorted(data_dict[keys[0]].keys())
        data = dict()
        for field in fields:
            data[field] = []
            for key in keys:
                data[field].append(data_dict[key][field])
        data["id"] = keys
        df = pd.DataFrame(data)
        df.to_csv(csv_path, index=False)

    dict_to_csv(train_data, train_path)
    dict_to_csv(valid_data, valid_path)
    dict_to_csv(test_data, test_paths[0])

    return train_path, valid_path, test_paths


def prepare_common_tokenizer(
    target_dir,
    cache_dir,
    tokenizer_data_path,
    get_path_only=False,
    tokenizer_name: str = None,
    vocab_file: str = None,
    vocab_type: str = "character",
    vocab_args: dict = None,
    slots_file: str = None,
):
    """
    Build the tokenizer following :obj:`SuperbASR.build_tokenizer` format

    Args:
        tokenizer_name (str): Save the tokenizer filepath with this filename
        vocab_file (str): When the tokenizer was already prepared, and just want
            to load and return the tokenizer here. Path or URL
        vocab_type (str): character / phoneme / word / subword
        vocab_args (dict):
            when :code:`vocab_type` is character / phoneme / word, supports arguments in
                :obj:`s3prl.dataio.encoder.vocabulary.generate_basic_vocab`

            whe :code:`vocab_type` is subword, supports arguments in
                :obj:`s3prl.dataio.encoder.vocabulary.generate_subword_vocab`
        slots_file (str): If presented, the pre-defined slots will be used to encode the
            special tokens. Path or URL

    Return:
        str

        tokenizer path
    """
    if tokenizer_name is None:
        tokenizer_name = f"{Path(tokenizer_data_path).stem}-{vocab_type}.tokenizer"
    tokenizer_path = Path(target_dir) / f"{tokenizer_name}.pkl"

    if get_path_only:
        return tokenizer_path

    if vocab_file is not None:
        vocab_file = str(vocab_file)
        if vocab_file.startswith("http"):
            vocab_file = urls_to_filepaths(vocab_file)

    if slots_file is not None:
        slots_file = str(slots_file)
        if slots_file.startswith("http"):
            slots_file = urls_to_filepaths(slots_file)

    if vocab_file is not None:
        tokenizer = load_tokenizer(
            vocab_type,
            vocab_file=vocab_file,
            slots_file=slots_file,
        )
    else:
        vocab_args = vocab_args or {}
        assert isinstance(vocab_args, dict)

        if vocab_type == "subword" and not "output_file" in vocab_args:
            vocab_args["output_file"] = Path(target_dir) / "tokenizer.spm"

        vocab_result = generate_vocab(
            vocab_type, text_file=str(tokenizer_data_path), **vocab_args
        )
        vocab_list = vocab_result if isinstance(vocab_result, list) else None
        vocab_file = vocab_result if isinstance(vocab_result, str) else None
        tokenizer = load_tokenizer(
            vocab_type,
            vocab_file=vocab_file,
            vocab_list=vocab_list,
            slots_file=slots_file,
        )

    with open(tokenizer_path, "wb") as f:
        pickle.dump(tokenizer, f)

    return tokenizer_path


class SuperbASR(ASR):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
                train_sets=["train-clean-100"],
                valid_sets=["dev-clean"],
                test_sets=["test-clean"],
            ),
            prepare_tokenizer_data=dict(),
            build_tokenizer=dict(
                vocab_type="character",
            ),
            build_dataset=dict(),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=32,
                    max_length=2000,  # due to this tiny max_length, the effective batch_size is always 16
                    shuffle=True,
                ),
                valid=dict(
                    batch_size=1,
                ),
                test=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                model_conf=dict(
                    module="LSTM",
                    proj_size=1024,
                    hidden_size=[1024, 1024],
                    dropout=[0.2, 0.2],
                    layer_norm=[False, False],
                    proj=[False, False],
                    sample_rate=[1, 1],
                    sample_style="concat",
                    bidirectional=True,
                ),
                specaug_conf=dict(
                    freq_mask_width_range=(0, 50),
                    num_freq_mask=4,
                    time_mask_width_range=(0, 40),
                    num_time_mask=2,
                ),
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                log_metrics=["cer", "wer"],
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-4,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(
                extra_conf=dict(
                    build_downstream_conf="${build_downstream}"
                ),  # This is redundant for ASR. Just to show how to clone other fields
            ),
            save_task=dict(),
            train=dict(
                total_steps=200000,
                log_step=100,
                eval_step=2000,
                save_step=500,
                gradient_clipping=1.0,
                gradient_accumulate=1,
                valid_metric="wer",
                valid_higher_better=False,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        """
        Prepare the task-specific data metadata (path, labels...).
        By default call :obj:`prepare_librispeech` with :code:`**prepare_data`

        Args:
            prepare_data (dict): same in :obj:`default_config`, support arguments in :obj:`prepare_librispeech`
            target_dir (str): Parse your corpus and save the csv file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            tuple

            1. train_path (str)
            2. valid_path (str)
            3. test_paths (List[str])

            Each path (str) should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            id                    (str) - the unique id for this data point
            wav_path              (str) - the absolute path of the waveform file
            transcription         (str) - a text string
            ====================  ====================
        """
        return prepare_librispeech(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )

    def prepare_tokenizer_data(
        self,
        prepare_tokenizer_data: dict,
        target_dir: str,
        cache_dir: str,
        train_csv: str,
        valid_csv: str,
        test_csvs: List[str],
        get_path_only: bool = False,
    ):
        """
        Prepare the text file used for training tokenizer.
        By default only use the transcription in the :code:`train_csv` returned from :obj:`prepare_data`
        The default :code:`prepare_tokenizer_data` build the character-based tokenizer

        Args:
            prepare_tokenizer_data (dict): same in :obj:`default_config`, no supported argument for now
            target_dir (str): Save the text file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            train_csv (str): The train data given by :obj:`prepare_data`
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            str

            The text file path, the text file should be in the format

            .. code-block:: none

                This is the first line
                This is the second line
                These are all text used for training tokenizer

        """
        tokenizer_data_name = f"{Path(train_csv).stem}.tokenizer_data"
        tokenizer_data_path = Path(target_dir) / f"{tokenizer_data_name}.txt"

        if get_path_only:
            return tokenizer_data_path

        all_text = pd.read_csv(train_csv)["transcription"]

        with tokenizer_data_path.open("w") as f:
            f.writelines([f"{line}\n" for line in all_text])

        return tokenizer_data_path

    def build_tokenizer(
        self,
        build_tokenizer: dict,
        target_dir: str,
        cache_dir: str,
        tokenizer_data_path: str,
        get_path_only: bool = False,
    ):
        """
        Build the tokenizer from the data prepared by :obj:`prepare_tokenizer_data`
        By default call :obj:`prepare_common_tokenizer` with :code:`**build_tokenizer`

        Args:
            build_tokenizer (dict): same in :obj:`default_config`, arguments for :obj:`prepare_common_tokenizer`
            target_dir (str): Current experinment directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            tokenizer_data_path (str): The text file from :obj:`prepare_tokenizer_data`
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            str

            filepath of the pickled :obj:`s3prl.dataio.encoder.tokenizer.Tokenizer`
        """
        return prepare_common_tokenizer(
            **self._get_current_arguments(flatten_dict="build_tokenizer")
        )

    def build_dataset(
        self,
        build_dataset: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        tokenizer_path: str,
    ):
        """
        Build the dataset for train/valid/test.

        Args:
            build_dataset (dict): same in :obj:`default_config`, not used
            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): The metadata csv file for the specific :code:`mode`
            tokenizer_path (str): The pickled tokenizer path for encoding transcription

        Returns:
            torch Dataset

            For all train/valid/test mode, the dataset should return each item as a dictionary
            containing the following keys:

            ====================  ====================
            key                   description
            ====================  ====================
            x                     (torch.FloatTensor) - the waveform in (seq_len, 1)
            x_len                 (int) - the waveform length :code:`seq_len`
            class_ids             (torch.LongTensor) - the encoded class ids of a transcription (sentence)
            labels                (str) - the text transcription
            unique_name           (str) - the unique id for this datapoint
            ====================  ====================
        """
        csv = pd.read_csv(data_csv)

        audio_loader = LoadAudio(csv["wav_path"].tolist())

        with open(tokenizer_path, "rb") as f:
            tokenizer = pickle.load(f)

        text_encoder = EncodeText(csv["transcription"].tolist(), tokenizer)
        ids = csv["id"].tolist()

        class Speech2TextDataset:
            def __len__(self):
                return len(audio_loader)

            def __getitem__(self, index: int):
                audio = audio_loader[index]
                text = text_encoder[index]
                return {
                    "x": audio["wav"],
                    "x_len": audio["wav_len"],
                    "class_ids": text["class_ids"],
                    "labels": text["labels"],
                    "unique_name": ids[index],
                }

        dataset = Speech2TextDataset()
        return dataset

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset: Dataset,
    ):
        """
        Return the batch sampler for torch DataLoader.

        Args:
            build_batch_sampler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                train                 (dict) - arguments for :obj:`SortedBucketingSampler`
                valid                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                test                  (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                ====================  ====================

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): the :code:`mode` specific csv from :obj:`prepare_data`
            dataset: the dataset from :obj:`build_dataset`

        Returns:
            batch sampler for torch DataLoader
        """

        @dataclass
        class Config:
            train: dict = None
            valid: dict = None
            test: dict = None

        conf = Config(**build_batch_sampler)

        if mode == "train":
            wav_lens = get_info(dataset, ["x_len"], Path(target_dir) / "train_stats")
            sampler = SortedBucketingSampler(wav_lens, **(conf.train or {}))
        elif mode == "valid":
            sampler = FixedBatchSizeBatchSampler(dataset, **(conf.valid or {}))
        elif mode == "test":
            sampler = FixedBatchSizeBatchSampler(dataset, **(conf.test or {}))

        return sampler

    def build_downstream(
        self,
        build_downstream: dict,
        downstream_input_size: int,
        downstream_output_size: int,
        downstream_input_stride: int,
    ):
        """
        Return the task-specific downstream model.
        By default build the :obj:`RNNEncoder` model wrapped with :obj:`ModelWithSpecaug`

        Args:
            build_downstream (dict): same in :obj:`default_config`, has two keys:
                :code:`model_conf` is the arguments for :obj:`RNNEncoder`;
                :code:`specaug_conf` is the arguments for :obj:`ModelWithSpecaug`
            downstream_input_size (int): the required input size of the model
            downstream_output_size (int): the required output size of the model
            downstream_input_stride (int): the input feature's stride (from 16 KHz)

        Returns:
            :obj:`s3prl.nn.interface.AbsFrameModel`
        """

        @dataclass
        class Config:
            model_conf: dict = None
            specaug_conf: dict = None

        conf = Config(**build_downstream)
        model = RNNEncoder(
            downstream_input_size, downstream_output_size, **(conf.model_conf or {})
        )
        downstream = ModelWithSpecaug(model, **(conf.specaug_conf or {}))
        return downstream



"""
Parse the Google Speech Commands V1 corpus

Authors:
  * Leo 2022
  * Cheng Liang 2022
"""

import hashlib
import logging
import re
from collections import OrderedDict
from pathlib import Path
from typing import List, Tuple, Union

from .base import Corpus

logger = logging.getLogger(__name__)

CLASSES = [
    "yes",
    "no",
    "up",
    "down",
    "left",
    "right",
    "on",
    "off",
    "stop",
    "go",
    "_unknown_",
    "_silence_",
]

__all__ = [
    "SpeechCommandsV1",
]


class SpeechCommandsV1(Corpus):
    """
    Args:
        dataset_root (str): should contain a 'dev' sub-folder for the training/validation set
            and a 'test' sub-folder for the testing set
    """

    def __init__(self, gsc1: str, gsc1_test: str, n_jobs: int = 4) -> None:
        train_dataset_root = Path(gsc1)
        test_dataset_root = Path(gsc1_test)

        train_list, valid_list = self.split_dataset(train_dataset_root)
        train_list = self.parse_train_valid_data_list(train_list, train_dataset_root)
        valid_list = self.parse_train_valid_data_list(valid_list, train_dataset_root)
        test_list = self.parse_test_data_list(test_dataset_root)

        self.train = self.list_to_dict(train_list)
        self.valid = self.list_to_dict(valid_list)
        self.test = self.list_to_dict(test_list)

        self._data = OrderedDict()
        self._data.update(self.train)
        self._data.update(self.valid)
        self._data.update(self.test)

    @staticmethod
    def split_dataset(
        root_dir: Union[str, Path], max_uttr_per_class=2**27 - 1
    ) -> Tuple[List[Tuple[str, str]], List[Tuple[str, str]]]:
        """Split Speech Commands into 3 set.

        Args:
            root_dir: speech commands dataset root dir
            max_uttr_per_class: predefined value in the original paper

        Return:
            train_list: [(class_name, audio_path), ...]
            valid_list: as above
        """
        train_list, valid_list = [], []

        for entry in Path(root_dir).iterdir():
            if not entry.is_dir() or entry.name == "_background_noise_":
                continue

            for audio_path in entry.glob("*.wav"):
                speaker_hashed = re.sub(r"_nohash_.*$", "", audio_path.name)
                hashed_again = hashlib.sha1(speaker_hashed.encode("utf-8")).hexdigest()
                percentage_hash = (int(hashed_again, 16) % (max_uttr_per_class + 1)) * (
                    100.0 / max_uttr_per_class
                )

                if percentage_hash < 10:
                    valid_list.append((entry.name, audio_path))
                elif percentage_hash < 20:
                    pass  # testing set is discarded
                else:
                    train_list.append((entry.name, audio_path))

        return train_list, valid_list

    @staticmethod
    def parse_train_valid_data_list(data_list, train_dataset_root: Path):
        data = [
            (class_name, audio_path)
            if class_name in CLASSES
            else ("_unknown_", audio_path)
            for class_name, audio_path in data_list
        ]
        data += [
            ("_silence_", audio_path)
            for audio_path in Path(train_dataset_root, "_background_noise_").glob(
                "*.wav"
            )
        ]
        return data

    @staticmethod
    def parse_test_data_list(test_dataset_root: Path):
        data = [
            (class_dir.name, audio_path)
            for class_dir in Path(test_dataset_root).iterdir()
            if class_dir.is_dir()
            for audio_path in class_dir.glob("*.wav")
        ]
        return data

    @staticmethod
    def path_to_unique_name(path: str):
        return "/".join(Path(path).parts[-2:])

    @classmethod
    def list_to_dict(cls, data_list):
        data = dict(
            {
                cls.path_to_unique_name(audio_path): {
                    "wav_path": audio_path,
                    "class_name": class_name,
                }
                for class_name, audio_path in data_list
            }
        )
        return data

    @property
    def all_data(self):
        """
        Return:
            Container: id (str)
                wav_path (str)
                class_name (str)
        """
        return self._data

    @property
    def data_split_ids(self):
        return list(self.train.keys()), list(self.valid.keys()), list(self.test.keys())

    @classmethod
    def download_dataset(cls, tgt_dir: str) -> None:
        import os
        import tarfile

        import requests

        assert os.path.exists(
            os.path.abspath(tgt_dir)
        ), "Target directory does not exist"

        def unzip_targz_then_delete(filepath: str, filename: str):
            file_path = os.path.join(
                os.path.abspath(tgt_dir), "CORPORA_DIR", filename.replace(".tar.gz", "")
            )
            os.makedirs(file_path)

            with tarfile.open(os.path.abspath(filepath)) as tar:
                tar.extractall(path=os.path.abspath(file_path))
            os.remove(os.path.abspath(filepath))

        def download_from_url(url: str):
            filename = url.split("/")[-1].replace(" ", "_")
            filepath = os.path.join(tgt_dir, filename)

            r = requests.get(url, stream=True)
            if r.ok:
                logger.info(f"Saving {filename} to", os.path.abspath(filepath))
                with open(filepath, "wb") as f:
                    for chunk in r.iter_content(chunk_size=1024 * 1024 * 10):
                        if chunk:
                            f.write(chunk)
                            f.flush()
                            os.fsync(f.fileno())
                logger.info(f"{filename} successfully downloaded")
                unzip_targz_then_delete(filepath, filename)
            else:
                logger.info(f"Download failed: status code {r.status_code}\n{r.text}")

        if not os.path.exists(
            os.path.join(os.path.abspath(tgt_dir), "CORPORA_DIR/speech_commands_v0.01")
        ):
            download_from_url(
                "http://download.tensorflow.org/data/speech_commands_v0.01.tar.gz"
            )
        if not os.path.exists(
            os.path.join(
                os.path.abspath(tgt_dir), "CORPORA_DIR/speech_commands_test_set_v0.01"
            )
        ):
            download_from_url(
                "http://download.tensorflow.org/data/speech_commands_test_set_v0.01.tar.gz"
            )
        logger.info(
            f"Speech commands dataset downloaded. Located at {os.path.abspath(tgt_dir)}/CORPORA_DIR/"
        )



import copy
import random
from dataclasses import dataclass

import torch

from .base import AugmentedDynamicItemDataset, DataPipe


@dataclass
class NoiseAugmentation(DataPipe):
    noise_proportion: float = 0.0
    input_feat_name: str = "input_feat"
    output_feat_name: str = "output_feat"
    """
    Args:
        noise_proportion (float): for this percentage of the time, Gaussian noise will be applied on all frames during MAM training, set to 0 for no noise
        input_feat_name (str): handle for the `takes` (input)
        output_feat_name (str): handle for the `provides` (output)
    """

    def apply_noise_on_data(self, input_feat):
        with torch.no_grad():
            if self.noise_proportion > 0:
                noised_feat = copy.deepcopy(input_feat)
                dice = random.random()
                if dice < self.noise_proportion:
                    noise_sampler = torch.distributions.Normal(0, 0.2)
                    noised_feat += noise_sampler.sample(noised_feat.shape)
                noised_feat = noised_feat.to(dtype=torch.float32)
                return noised_feat
            else:
                return input_feat

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        dataset.add_dynamic_item(
            self.apply_noise_on_data,
            takes=self.input_feat_name,
            provides=self.output_feat_name,
        )
        return dataset



import pytest
from dotenv import dotenv_values

from s3prl.dataio.corpus.snips import SNIPS


@pytest.mark.corpus
def test_snips():
    config = dotenv_values()
    dataset_root = config["SNIPS"]
    dataset = SNIPS(
        dataset_root,
        [
            "Ivy",
            "Joanna",
            "Joey",
            "Justin",
            "Kendra",
            "Kimberly",
            "Matthew",
            "Salli",
        ],
        ["Aditi", "Amy", "Geraint", "Nicole"],
        ["Brian", "Emma", "Raveena", "Russell"],
    )
    train_data, valid_data, test_data = dataset.data_split

    assert len(train_data) == 104672
    assert len(valid_data) == 2800
    assert len(test_data) == 2800



from __future__ import annotations

import logging
from typing import List

import torch

from s3prl import Logs, Output
from s3prl.nn.predictor_mockingjay import PredictorMockingjay as PredictorExample
from s3prl.nn.transformer_mockingjay import TransformerMockingjay as UpstreamExample

from . import Task

logger = logging.getLogger(__name__)


class FeatReconstructionTask(Task):
    """
    Attributes:
        upstream (torch.nn.Module): The upstream encoder (transformers, rnn, etc) that outputs `hidden_states`
        predictor (torch.nn.Module): The pre-training predictor that takes `hidden_states` as input and maps to the task target
        loss (torch.nn Loss Functions): The reconstruction loss (torch.nn.L1Loss, torch.nn.MSELoss, etc)
    """

    def __init__(
        self,
        upstream: UpstreamExample,
        predictor: PredictorExample,
        loss: torch.nn.L1Loss,
        loss_config: dict = {},
        **kwargs,
    ):
        """
        The input feature does not necessary have to be the same as the target feature.

        Args:
            upstream (Encoder)
            predictor (Projection NN)
            loss (reconstruction loss)
                feat_A -> upstream -> predictor -> feat_B
                loss(feat_A, feat_B)
        """

        super().__init__()
        self.upstream = upstream
        self.predictor = predictor
        self.loss = loss(**loss_config)

    def predict(
        self,
        x: torch.Tensor,
        label: torch.Tensor,
        label_mask: torch.BoolTensor = None,
        position_encoding: torch.Tensor = None,
        attention_mask: torch.LongTensor = None,
    ):
        """
        Args:
            x (torch.Tensor): source_feat - (batch_size, timestamps, input_size)
            label (torch.Tensor): target_feat - (batch_size, timestamps, output_size)
            label_mask (torch.BoolTensor): (batch_size, timestamps, output_size)
            position_encoding (torch.Tensor): (batch_size, timestamps, input_size)
            attention_mask (torch.LongTensor): (batch_size, timestamps)

        Return:
            hidden_states (torch.Tensor): (batch_size, timestamps, hidden_size)
            loss (torch.Tensor): scalar.
            prediction (torch.Tensor): (batch_size, timestamps, output_size)
        """
        if position_encoding is None and attention_mask is None:
            upstream_output: torch.Tensor = self.upstream(x)
        else:
            upstream_output: torch.Tensor = self.upstream(
                x, position_encoding, attention_mask
            )
        prediction: torch.Tensor = self.predictor(upstream_output).prediction

        if label_mask is None:
            reconstruction_loss = self.loss(prediction, label)
        else:
            assert label_mask.sum() > 0, "Without any masking, loss might go NaN."
            reconstruction_loss = self.loss(
                prediction.masked_select(label_mask), label.masked_select(label_mask)
            )

        return Output(
            loss=reconstruction_loss,
            hidden_states=upstream_output.hidden_states,
            prediction=prediction,
        )

    def _general_forward(
        self,
        x: torch.Tensor,
        label: torch.Tensor,
        label_mask: torch.BoolTensor = None,
        position_encoding: torch.Tensor = None,
        attention_mask: torch.LongTensor = None,
        unique_name: List[str] = None,
    ):
        loss, hidden_states, prediction = self.predict(
            x, label, label_mask, position_encoding, attention_mask
        ).slice(3)

        logs = Logs()
        logs.add_hidden_state("hidden_states", hidden_states)
        logs.add_hidden_state("prediction", prediction)

        return Output(
            loss=loss,
            prediction=prediction,
            label=label,
            unique_name=unique_name,
            logs=logs,
        )

    def _general_reduction(self, batch_results: list, on_epoch_end: bool = None):
        losses = []
        for batch_result in batch_results:
            losses.append(batch_result.loss)

        loss = (sum(losses) / len(losses)).item()

        logs = Logs()
        logs.add_scalar("loss", loss)

        return Output(
            logs=logs,
        )

    def train_step(
        self,
        x: torch.Tensor,
        label: torch.Tensor,
        label_mask: torch.BoolTensor = None,
        position_encoding: torch.Tensor = None,
        attention_mask: torch.LongTensor = None,
        unique_name: List[str] = None,
        **kwargs,
    ):
        """
        Each forward step in the training loop

        Args:
            source_feat (torch.Tensor): (batch_size, timestamps, input_size)
            target_feat (torch.Tensor): (batch_size, timestamps, output_size)
            label_mask (torch.BoolTensor): (batch_size, timestamps, output_size)
            pos_enc (torch.Tensor): (batch_size, timestamps, input_size)
            attn_mask (torch.LongTensor): (batch_size, timestamps)

        Return:
            hidden_states (torch.Tensor): (batch_size, timestamps, hidden_size)
            loss (torch.Tensor): scalar.
            prediction (torch.Tensor): (batch_size, timestamps, output_size)
        """
        return self._general_forward(
            x, label, label_mask, position_encoding, attention_mask, unique_name
        )

    def train_reduction(self, batch_results: list, on_epoch_end: bool = False):
        """
        After several forward steps, outputs should be collected untouched (but detaching the Tensors)
        into a list and passed as batch_results. This function examine the collected items and compute
        metrics across these batches. This function might be called in the middle of an epoch for quick
        logging, or after exactly an epoch to know the epoch level performance.

        Args:
            batch_results (List[cacheable version of the output of self.train_step])
            on_epoch_end (bool):
                usually you should keep the same behavior between sub-epoch and epoch level
                this parameter is here in case you need specific postprocessing which must
                only be done right on the end of an epoch

        Return:
            logs (List[Log]):
                a list of content to log onto any logger
                each content should be in the Log class format
        """
        return self._general_reduction(batch_results, on_epoch_end)

    def valid_step(
        self,
        x: torch.Tensor,
        label: torch.Tensor,
        label_mask: torch.BoolTensor = None,
        position_encoding: torch.Tensor = None,
        attention_mask: torch.LongTensor = None,
        unique_name: List[str] = None,
        **kwargs,
    ):
        return self._general_forward(
            x, label, label_mask, position_encoding, attention_mask, unique_name
        )

    def test_step(
        self,
        x: torch.Tensor,
        label: torch.Tensor,
        label_mask: torch.BoolTensor = None,
        position_encoding: torch.Tensor = None,
        attention_mask: torch.LongTensor = None,
        unique_name: List[str] = None,
        **kwargs,
    ):
        return self._general_forward(
            x, label, label_mask, position_encoding, attention_mask, unique_name
        )

    def valid_reduction(self, batch_results: list, on_epoch_end: bool = True):
        return self._general_reduction(batch_results, on_epoch_end)

    def test_reduction(self, batch_results: list, on_epoch_end: bool = True):
        return self._general_reduction(batch_results, on_epoch_end)



import logging

import torch
from torch.nn.utils.rnn import pad_sequence

SAMPLE_RATE = 16000
EXAMPLE_SEC = 5

logger = logging.getLogger(__name__)

try:
    from espnet2.tasks.hubert import HubertTask
except ModuleNotFoundError:
    HubertTask = None
    logger.warning("ESPnet is not installed, cannot use espnet_hubert upstream")


class UpstreamExpert(torch.nn.Module):
    def __init__(self, ckpt, config=None, **kwargs):
        super().__init__()
        device = "cuda" if torch.cuda.is_available() else "cpu"
        assert (
            HubertTask is not None
        ), "ESPnet is not installed, run `external_tools/install_espnet.sh` to install"
        hubert_model, hubert_train_args = HubertTask.build_model_from_file(
            config,
            ckpt,
            device,
        )
        self.device = next(hubert_model.parameters()).device
        self.model = hubert_model.encoder.hubert_pretrain_model

    def get_downsample_rates(self, key: str = None) -> int:
        return 320

    def forward(self, wavs):
        wav_lengths = torch.LongTensor([len(wav) for wav in wavs]).to(self.device)
        wavs = pad_sequence(wavs, batch_first=True).to(self.device)
        feats = self.model.wav2vec2.extract_features(wavs, wav_lengths)[
            0
        ]  # (time, feat_dim))

        return {"hidden_states": feats}



import torch.nn as nn

from s3prl import Output


class PredictorIdentity(nn.Module):
    """
    This nn module is used as a predictor placeholder for certain SSL problems.
    """

    def __init__(self, **kwargs):
        super(PredictorIdentity, self).__init__()

    def forward(self, output: Output):
        """
        Args:
            output (s3prl.Output): An Output module

        Return:
            output (s3prl.Output): exactly the same as input, an Output module
        """
        return output



"""
Parse the LibriSpeech corpus

Authors:
  * Heng-Jui Chang 2022
"""

import logging
import os
import re
from collections import OrderedDict, defaultdict
from pathlib import Path
from typing import Any, Dict, List

from joblib import Parallel, delayed

from .base import Corpus

logger = logging.getLogger(__name__)

LIBRI_SPLITS = [
    "train-clean-100",
    "train-clean-360",
    "train-other-500",
    "dev-clean",
    "dev-other",
    "test-clean",
    "test-other",
]

__all__ = [
    "LibriSpeech",
]


def read_text(file: Path) -> str:
    src_file = "-".join(str(file).split("-")[:-1]) + ".trans.txt"
    idx = file.stem.replace(".flac", "")

    with open(src_file, "r") as fp:
        for line in fp:
            if idx == line.split(" ")[0]:
                return line[:-1].split(" ", 1)[1]

    logger.warning(f"Transcription of {file} not found!")


def check_no_repeat(splits: List[str]) -> bool:
    count = defaultdict(int)
    for split in splits:
        count[split] += 1

    repeated = ""
    for key, val in count.items():
        if val > 1:
            repeated += f" {key} ({val} times)"

    if len(repeated) != 0:
        logger.warning(
            f"Found repeated splits in corpus: {repeated}, which might cause unexpected behaviors."
        )
        return False

    return True


def _parse_spk_to_gender(speaker_file: Path) -> dict:
    speaker_file = Path(speaker_file)
    with speaker_file.open() as file:
        lines = [line.strip() for line in file.readlines()]
    for line_id in range(len(lines)):
        line = lines[line_id]
        if "SEX" in line and "SUBSET" in line and "MINUTES" in line and "NAME" in line:
            break

    line_id += 1  # first line with speaker info
    spk2gender = {}
    for line_id in range(line_id, len(lines)):
        line = lines[line_id]
        line = re.sub("\t+", " ", line)
        line = re.sub(" +", " ", line)
        parts = line.split("|", maxsplit=4)
        ID, SEX, SUBSET, MINUTES, NAME = parts
        spk2gender[int(ID)] = SEX.strip()
    return spk2gender


class LibriSpeech(Corpus):
    """LibriSpeech Corpus
    Link: https://www.openslr.org/12

    Args:
        dataset_root (str): Path to LibriSpeech corpus directory.
        n_jobs (int, optional): Number of jobs. Defaults to 4.
        train_split (List[str], optional): Training splits. Defaults to ["train-clean-100"].
        valid_split (List[str], optional): Validation splits. Defaults to ["dev-clean"].
        test_split (List[str], optional): Testing splits. Defaults to ["test-clean"].
    """

    def __init__(
        self,
        dataset_root: str,
        n_jobs: int = 4,
        train_split: List[str] = ["train-clean-100"],
        valid_split: List[str] = ["dev-clean"],
        test_split: List[str] = ["test-clean"],
    ) -> None:
        self.dataset_root = Path(dataset_root).resolve()
        self.train_split = train_split
        self.valid_split = valid_split
        self.test_split = test_split
        self.all_splits = train_split + valid_split + test_split
        assert check_no_repeat(self.all_splits)

        self.data_dict = self._collect_data(dataset_root, self.all_splits, n_jobs)
        self.train = self._data_to_dict(self.data_dict, train_split)
        self.valid = self._data_to_dict(self.data_dict, valid_split)
        self.test = self._data_to_dict(self.data_dict, test_split)

        self._data = OrderedDict()
        self._data.update(self.train)
        self._data.update(self.valid)
        self._data.update(self.test)

    def get_corpus_splits(self, splits: List[str]):
        return self._data_to_dict(self.data_dict, splits)

    @property
    def all_data(self):
        """
        Return all the data points in a dict of the format

        .. code-block:: yaml

            data_id1:
                wav_path: (str) The waveform path
                transcription: (str) The transcription
                speaker: (str) The speaker name
                gender: (str) The speaker's gender
                corpus_split: (str) The split of corpus this sample belongs to

            data_id2:
                ...
        """
        return self._data

    @property
    def data_split_ids(self):
        return (
            list(self.train.keys()),
            list(self.valid.keys()),
            list(self.test.keys()),
        )

    @staticmethod
    def _collect_data(
        dataset_root: str, splits: List[str], n_jobs: int = 4
    ) -> Dict[str, Dict[str, List[Any]]]:
        spkr2gender = _parse_spk_to_gender(Path(dataset_root) / "SPEAKERS.TXT")
        data_dict = {}
        for split in splits:
            split_dir = os.path.join(dataset_root, split)
            if not os.path.exists(split_dir):
                logger.info(f"Split {split} is not downloaded. Skip data collection.")
                continue

            wav_list = list(Path(split_dir).rglob("*.flac"))
            name_list = [file.stem.replace(".flac", "") for file in wav_list]
            text_list = Parallel(n_jobs=n_jobs)(
                delayed(read_text)(file) for file in wav_list
            )
            spkr_list = [int(name.split("-")[0]) for name in name_list]

            wav_list, name_list, text_list, spkr_list = zip(
                *[
                    (wav, name, text, spkr)
                    for (wav, name, text, spkr) in sorted(
                        zip(wav_list, name_list, text_list, spkr_list),
                        key=lambda x: x[1],
                    )
                ]
            )
            data_dict[split] = {
                "name_list": list(name_list),
                "wav_list": list(wav_list),
                "text_list": list(text_list),
                "spkr_list": list(spkr_list),
                "gender_list": [spkr2gender[spkr] for spkr in spkr_list],
            }

        return data_dict

    @staticmethod
    def _data_to_dict(
        data_dict: Dict[str, Dict[str, List[Any]]], splits: List[str]
    ) -> dict():
        data = dict(
            {
                name: {
                    "wav_path": data_dict[split]["wav_list"][i],
                    "transcription": data_dict[split]["text_list"][i],
                    "speaker": data_dict[split]["spkr_list"][i],
                    "gender": data_dict[split]["gender_list"][i],
                    "corpus_split": split,
                }
                for split in splits
                for i, name in enumerate(data_dict[split]["name_list"])
            }
        )
        return data

    @classmethod
    def download_dataset(
        cls,
        target_dir: str,
        splits: List[str] = ["train-clean-100", "dev-clean", "test-clean"],
    ) -> None:
        import os
        import tarfile

        import requests

        target_dir = Path(target_dir)
        target_dir.mkdir(exist_ok=True, parents=True)

        def unzip_targz_then_delete(filepath: str):
            with tarfile.open(os.path.abspath(filepath)) as tar:
                tar.extractall(path=os.path.abspath(target_dir))
            os.remove(os.path.abspath(filepath))

        def download_from_url(url: str):
            filename = url.split("/")[-1].replace(" ", "_")
            filepath = os.path.join(target_dir, filename)

            r = requests.get(url, stream=True)
            if r.ok:
                logger.info(f"Saving {filename} to {os.path.abspath(filepath)}")
                with open(filepath, "wb") as f:
                    for chunk in r.iter_content(chunk_size=1024 * 1024 * 10):
                        if chunk:
                            f.write(chunk)
                            f.flush()
                            os.fsync(f.fileno())
                logger.info(f"{filename} successfully downloaded")
                unzip_targz_then_delete(filepath)
            else:
                logger.info(f"Download failed: status code {r.status_code}\n{r.text}")

        for split in splits:
            if not os.path.exists(
                os.path.join(os.path.abspath(target_dir), "Librispeech/" + split)
            ):
                download_from_url(
                    "https://www.openslr.org/resources/12/" + split + ".tar.gz"
                )
        logger.info(
            ", ".join(splits)
            + f"downloaded. Located at {os.path.abspath(target_dir)}/Librispeech/"
        )



import json
from pathlib import Path
from typing import List

import torchaudio
from joblib import Parallel, delayed
from tqdm import tqdm

_default_cache_dir = Path.home() / ".cache" / "s3prl" / "audio_info"


__all__ = [
    "get_cache_dir",
    "set_cache_dir",
    "get_audio_info",
]


def get_cache_dir():
    _default_cache_dir.mkdir(exist_ok=True, parents=True)
    return _default_cache_dir


def set_cache_dir(cache_dir: str):
    global _default_cache_dir
    _default_cache_dir = Path(cache_dir)


def get_audio_info(
    audio_paths: List[str],
    audio_ids: List[str],
    cache_dir: str = None,
    num_workers: int = 6,
) -> List[dict]:
    """
    Use :code:`torchaudio.info` to retrieve the metadata from audio paths.
    The retrieved metadata is cached in :code:`cache_dir`
    """

    cache_dir = cache_dir or get_cache_dir()
    cache_dir: Path = Path(cache_dir)

    def _get_info(audio_path: str, audio_id: str):
        cache_file = cache_dir / f"{audio_id}.json"
        if cache_file.is_file():
            with cache_file.open() as f:
                info = json.load(f)
                return info

        torchaudio.set_audio_backend("sox_io")
        torchaudio_info = torchaudio.info(audio_path)
        info = {
            "sample_rate": torchaudio_info.sample_rate,
            "num_frames": torchaudio_info.num_frames,
            "num_channels": torchaudio_info.num_channels,
            "bits_per_sample": torchaudio_info.bits_per_sample,
            "encoding": torchaudio_info.encoding,
        }
        return info

    infos = Parallel(n_jobs=num_workers)(
        delayed(_get_info)(path, idx)
        for path, idx in tqdm(zip(audio_paths, audio_ids), desc="Get audio metadata")
    )
    return infos
